import argparse
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os
import json
from collections import OrderedDict


import h5py
import hdf5plugin

import numpy as np
import modiscolite
import modiscolite.tfmodisco
import modiscolite.util

import random
import numpy

random.seed(0)
numpy.random.seed(0)

from tensorflow.random import set_seed 
set_seed(1234)


def save_hierarchy(h, grp):
	root_node_names = []
	for i in range(len(h.root_nodes)):
		node_name = "root_node"+str(i)
		root_node_names.append(node_name)
		save_node(h.root_nodes[i], grp.create_group(node_name))

	dset = grp.create_dataset("root_node_names", (len(root_node_names),), dtype=h5py.special_dtype(vlen=bytes))
	dset[:] = root_node_names


def save_node(node, grp):
	if node.indices_merged is not None:
		grp.create_dataset("indices_merged", data=np.array(node.indices_merged)) 
		grp.create_dataset("submat_crosscontam", data=np.array(node.submat_crosscontam)) 
		grp.create_dataset("submat_alignersim", data=np.array(node.submat_alignersim))

	save_pattern(node.pattern, grp.create_group("pattern"))

	if node.child_nodes is not None:
		child_node_names = []
		for i in range(len(node.child_nodes)):
			child_node_name = "child_node"+str(i)
			child_node_names.append(child_node_name)
			save_node(node.child_nodes[i], grp.create_group(child_node_name))
			#node.child_nodes[i].save_hdf5(grp.create_group(child_node_name))

		dset = grp.create_dataset("child_node_names", (len(child_node_names),), dtype=h5py.special_dtype(vlen=bytes))
		dset[:] = child_node_names


def save_pattern(pattern, grp):
	track_grp = grp.create_group("sequence")
	track_grp.create_dataset("fwd", data=pattern.sequence)
	track_grp.create_dataset("rev", data=pattern.sequence[::-1, ::-1])
	track_grp.attrs["has_pos_axis"] = True

	track_grp = grp.create_group("task0_contrib_scores")
	track_grp.create_dataset("fwd", data=pattern.contrib_scores)
	track_grp.create_dataset("rev", data=pattern.contrib_scores[::-1, ::-1])
	track_grp.attrs["has_pos_axis"] = True

	track_grp = grp.create_group("task0_hypothetical_contribs")
	track_grp.create_dataset("fwd", data=pattern.hypothetical_contribs)
	track_grp.create_dataset("rev", data=pattern.hypothetical_contribs[::-1, ::-1])
	track_grp.attrs["has_pos_axis"] = True

	seqlets_and_alnmts_grp = grp.create_group("seqlets_and_alnmts")

	dset = seqlets_and_alnmts_grp.create_dataset("seqlets", (len(pattern.seqlets),), dtype=h5py.special_dtype(vlen=bytes))
	dset[:] = [str(x) for x in pattern.seqlets]

	seqlets_and_alnmts_grp.create_dataset("alnmts", 
		data=np.zeros(len(pattern.seqlets)))

	if pattern.subclusters is not None:
		grp.create_dataset("subclusters", data=pattern.subclusters)
		subcluster_to_subpattern_grp = grp.create_group("subcluster_to_subpattern")

		dset = subcluster_to_subpattern_grp.create_dataset("subcluster_names", (len(pattern.subcluster_to_subpattern.keys()),), dtype=h5py.special_dtype(vlen=bytes))
		dset[:] = ["subcluster_"+str(x) for x in pattern.subcluster_to_subpattern.keys()]

		for subcluster, subpattern in pattern.subcluster_to_subpattern.items():
			subpattern_grp = subcluster_to_subpattern_grp.create_group("subcluster_"+str(subcluster)) 
			save_pattern(subpattern, subpattern_grp)


def save_patterns(patterns, grp):
	all_pattern_names = []
	for idx, pattern in enumerate(patterns):
		pattern_name = "pattern_"+str(idx)
		all_pattern_names.append(pattern_name)
		pattern_grp = grp.create_group(pattern_name)
		save_pattern(pattern, pattern_grp)

	dset = grp.create_dataset("all_pattern_names", (len(all_pattern_names),), dtype=h5py.special_dtype(vlen=bytes))
	dset[:] = all_pattern_names

def save_string_list(string_list, dset_name, grp):
	dset = grp.create_dataset(dset_name, (len(string_list),),
							  dtype=h5py.special_dtype(vlen=bytes))
	dset[:] = string_list

def save_seqlet_coords(seqlets, dset_name, grp):
	coords_strings = [str(x) for x in seqlets] 
	save_string_list(string_list=coords_strings,
					 dset_name=dset_name, grp=grp)

def motif_discovery_main(scores_path, output_directory, modisco_window_size, max_seqlets):

	if not os.path.exists(scores_path):
		raise Exception(
			"Score file {} does not exist".format(scores_path))

	if not os.path.exists(output_directory):
		raise Exception(
			"Output directiry {} does not exist".format(output_directory))

	# Load the scores
	scores = h5py.File(scores_path, 'r')
	
	try:
		center = scores['hyp_scores'].shape[1] // 2
		start = center - modisco_window_size // 2
		end = center + modisco_window_size // 2
		
		shap_scores = scores['hyp_scores'][:, start:end, :]
		one_hot_seqs = scores['input_seqs'][:, start:end, :]
	except KeyError:
		center = scores['shap']['seq'].shape[2] // 2
		start = center - modisco_window_size // 2
		end = center + modisco_window_size // 2
		
		shap_scores = scores['shap']['seq'][:, :, start:end].transpose(0, 2, 1)
		one_hot_seqs = scores['raw']['seq'][:, :, start:end].transpose(0, 2, 1)
	
	proj_shap_scores = np.multiply(one_hot_seqs, shap_scores)

	scores.close()

	(multitask_seqlet_creation_results, metaclustering_results, 
		metacluster_idx_to_submetacluster_results) = modiscolite.tfmodisco.TFMoDISco(
		hypothetical_contribs=shap_scores, one_hot=one_hot_seqs,
		max_seqlets_per_metacluster=max_seqlets,
		sliding_window_size=20,
		flank_size=5,
		target_seqlet_fdr=0.05,
		n_leiden_runs=1)

	modisco_results_path = '{}/modisco_results.h5'.format(
		output_directory)
		

	########
	### Ugly saving code
	########

	grp = h5py.File(modisco_results_path, 'w')

	multitask_group = grp.create_group("multitask_seqlet_creation_results")
	save_string_list(
		  string_list=list(multitask_seqlet_creation_results['task_name_to_coord_producer_results'].keys()),
		  dset_name="task_names", grp=multitask_group)

	multitask2_group = multitask_group.create_group("multitask_seqlet_creator")
	multitask2_group.create_group("coord_producer")
	multitask2_group.create_group("overlap_resolver")
	multitask2_group.attrs['verbose'] = True

	save_seqlet_coords(seqlets=multitask_seqlet_creation_results['final_seqlets'], dset_name="final_seqlets", grp=multitask_group)

	tntcpg = multitask_group.create_group("task_name_to_coord_producer_results")

	coord_producer_results = multitask_seqlet_creation_results['task_name_to_coord_producer_results']['task0']
	save_string_list(string_list=[str(x) for x in coord_producer_results['seqlets']], dset_name="coords", grp=tntcpg)

	tnt_group = tntcpg.create_group("tnt_results")
	tnt_group.attrs['class'] = 'FWACTransformAndThresholdResults'

	metaclustering_group = grp.create_group("metaclustering_results")
	metaclustering_group.create_dataset("metacluster_indices", data=metaclustering_results['metacluster_indices'])

	metaclustering_group.attrs["class"] = "SignBasedPatternClustering"
	save_string_list(['task0'], dset_name="task_names", grp=metaclustering_group) 
	
	task_name_to_value_provider_grp = metaclustering_group.create_group("task_name_to_value_provider")

	task0_group = task_name_to_value_provider_grp.create_group("task0")
	task0_group.attrs["class"] = "TransformCentralWindowValueProvider"
	task0_group.attrs["track_name"] = "task0_contrib_scores"
	task0_group.attrs["central_window"] = 20
	
	val_transformer_group = task0_group.create_group("val_transformer")
	val_transformer_group.attrs["class"] = "AbsPercentileValTransformer"
	
	metaclustering_group.attrs["min_cluster_size"] = 100
	metaclustering_group.attrs["verbose"] = True
	metaclustering_group.attrs["fit_called"] = True

	activity_pattern_to_cluster_idx_grp = metaclustering_group.create_group("activity_pattern_to_cluster_idx")
	for activity_pattern, cluster_idx in metaclustering_results['pattern_to_cluster_idx'].items():
		activity_pattern_to_cluster_idx_grp.attrs[activity_pattern] = cluster_idx

	group = grp.create_group("metacluster_idx_to_submetacluster_results")
	for idx, results in metacluster_idx_to_submetacluster_results.items():
		smc_group = group.create_group("metacluster_{}".format(idx))
		smc_group.attrs['size'] = results['metacluster_size']
		save_seqlet_coords(seqlets=results['seqlets'], dset_name="seqlets", grp=smc_group)

		patterns = results['seqlets_to_patterns_result']
		pattern_group = smc_group.create_group("seqlets_to_patterns_result")
		pattern_group.attrs["success"] = patterns['success']
		pattern_group.attrs["other_config"] = json.dumps(patterns['other_config'], indent=4, separators=(',', ': '))

		if patterns['success']:
			if patterns['each_round_initcluster_motifs'] is not None:
				rounds_motifs = pattern_group.create_group("each_round_initcluster_motifs")
				all_round_names = []

				for round_idx, initcluster_motifs in enumerate(patterns['each_round_initcluster_motifs']):
					round_name = "round_" + str(round_idx)
					save_patterns(patterns=initcluster_motifs, grp=rounds_motifs.create_group(round_name)) 

				save_string_list(string_list=all_round_names, dset_name="all_round_names", grp=rounds_motifs)


			save_patterns(patterns['patterns'], pattern_group.create_group("patterns"))
			save_patterns(patterns['remaining_patterns'], pattern_group.create_group("remaining_patterns"))

			cluster_group = pattern_group.create_group("cluster_results")
			cluster_group.attrs['class'] = 'ClusterResults'
			cluster_group.create_dataset("cluster_indices", data=patterns['cluster_results']['cluster_indices'])

			save_hierarchy(patterns['pattern_merge_hierarchy'], pattern_group.create_group("pattern_merge_hierarchy"))

			#patterns['pattern_merge_hierarchy'].save_hdf5(grp=pattern_group.create_group("pattern_merge_hierarchy"))


	#tfmodisco_results.save_hdf5(h5py.File(modisco_results_path, 'w'))
	print("Saved modisco results to file {}".format(str(modisco_results_path)))
	####

	seqlet_path = '{}/seqlets.txt'.format(output_directory)
	#print("Saving seqlets to %s" % seqlet_path)
	seqlets = metacluster_idx_to_submetacluster_results[0]['seqlets']
	bases = np.array(["A", "C", "G", "T"])
	with open(seqlet_path, "w") as f:
		for seqlet in seqlets:
			sequence = "".join(bases[np.argmax(seqlet.sequence, axis=-1)])
			example_index = seqlet.example_idx
			start, end = seqlet.start, seqlet.end
			f.write(">example%d:%d-%d\n" % (example_index, start, end))
			f.write(sequence + "\n")

	#print("Saving pattern visualizations")

	patterns = (metacluster_idx_to_submetacluster_results[0]['seqlets_to_patterns_result']['patterns'])

	# generate .pngs of each motif and write motif seqlet to
	# individual files
	for idx,pattern in enumerate(patterns):		
		pattern_seqlet_path = os.path.join(output_directory,
										   'pattern{}_seqlets.txt'.format(idx))
		with open(pattern_seqlet_path, "w") as f: 
			for seqlet in pattern.seqlets:
				sequence = "".join(
					bases[np.argmax(seqlet.sequence, axis=-1)]
				)
				example_index = seqlet.example_idx
				start, end = seqlet.start, seqlet.end
				f.write(">example%d:%d-%d\n" % (example_index, start, end))
				f.write(sequence + "\n")



if __name__ == '__main__':
	n_seqlets = 100000

	exp = "ENCSR450BLH"
	exp = "GM12878_ATAC"
	scores_path = "{}/attributions/counts_scores.h5".format(exp)
	#output_directory = "large_runs/{}/reference_modisco/".format(n_seqlets)
	output_directory = "large_runs/{}/new_modisco_atac/".format(n_seqlets)

	motif_discovery_main(scores_path, output_directory, 400, n_seqlets)



# affinitymat.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import sklearn
import sklearn.manifold

import numpy as np

import scipy
from scipy.sparse import coo_matrix

from numba import njit
from numba import prange

from . import util
from . import gapped_kmer


@njit('float64(float64[:], int64[:], int64[:], float64[:], int64[:], int64[:], int64, int64)')
def _sparse_vv_dot(X_data, X_indices, X_indptr, Y_data, Y_indices, Y_indptr, i, j):
	xi = X_indptr[i]
	yj = Y_indptr[j]
	dot = 0.0

	while xi < X_indptr[i+1] and yj < Y_indptr[j+1]:
		x_col = X_indices[xi]
		x_data = X_data[xi]

		y_col = Y_indices[yj]
		y_data = Y_data[yj]

		if x_col == y_col:
			dot += x_data * y_data
			xi += 1
			yj += 1

		elif x_col < y_col:
			xi += 1

		else:
			yj += 1

	return dot

@njit(parallel=True)
def _sparse_mm_dot(X_data, X_indices, X_indptr, Y_data, Y_indices, Y_indptr, k):
	n_rows = len(Y_indptr) - 1

	neighbors = np.empty((n_rows, k), dtype='int32')
	sims = np.empty((n_rows, k), dtype='float64')

	for i in prange(n_rows):
		dot = np.zeros(n_rows, dtype='float64')

		for j in range(n_rows):
			xdot = _sparse_vv_dot(X_data, X_indices, X_indptr, X_data, X_indices, X_indptr, i, j)
			ydot = _sparse_vv_dot(X_data, X_indices, X_indptr, Y_data, Y_indices, Y_indptr, i, j)
			dot[j] = max(xdot, ydot)

		dot_argsort = np.argsort(-dot, kind='mergesort')[:k]
		neighbors[i] = dot_argsort
		sims[i] = dot[dot_argsort]

	return sims, neighbors

def cosine_similarity_from_seqlets(seqlets, n_neighbors, sign, topn=20, 
	min_k=4, max_k=6, max_gap=15, max_len=15, max_entries=500, 
	alphabet_size=4):

	X_fwd = gapped_kmer._seqlet_to_gkmers(seqlets, topn, 
		min_k, max_k, max_gap, max_len, max_entries, True, sign)

	X_bwd = gapped_kmer._seqlet_to_gkmers(seqlets, topn, min_k, max_k, max_gap, 
			max_len, max_entries, False, sign)

	X = sklearn.preprocessing.normalize(X_fwd, norm='l2', axis=1)
	Y = sklearn.preprocessing.normalize(X_bwd, norm='l2', axis=1)

	n, d = X.shape
	k = min(n_neighbors+1, n)
	return _sparse_mm_dot(X.data, X.indices, X.indptr, Y.data, Y.indices, Y.indptr, k)


def jaccard_from_seqlets(seqlets, min_overlap, filter_seqlets=None, 
	seqlet_neighbors=None):

	all_fwd_data, all_rev_data = util.get_2d_data_from_patterns(seqlets)

	if filter_seqlets is None:
		filter_seqlets = seqlets
		filters_all_fwd_data = all_fwd_data
		filters_all_rev_data = all_rev_data
	else:
		filters_all_fwd_data, filters_all_rev_data = util.get_2d_data_from_patterns(filter_seqlets)

	if seqlet_neighbors is None:
		seqlet_neighbors = [list(range(len(filter_seqlets)))
							for x in seqlets] 

	#apply the cross metric
	affmat_fwd = jaccard(seqlet_neighbors=seqlet_neighbors, 
		X=filters_all_fwd_data,
		Y=all_fwd_data, min_overlap=min_overlap, func=int, 
		return_sparse=True)

	affmat_rev = jaccard(seqlet_neighbors=seqlet_neighbors,
		X=filters_all_rev_data, Y=all_fwd_data,
		min_overlap=min_overlap, func=int,
		return_sparse=True) 

	affmat = np.maximum(affmat_fwd, affmat_rev)
	return affmat


def jaccard(X, Y, min_overlap=None, seqlet_neighbors=None, func=np.ceil, 
	return_sparse=False):

	if seqlet_neighbors is None:
		seqlet_neighbors = np.tile(np.arange(X.shape[0]), (Y.shape[0], 1))

	if min_overlap is not None:
		n_pad = int(func(X.shape[1]*(1-min_overlap)))
		pad_width = ((0,0), (n_pad, n_pad), (0,0)) 
		Y = np.pad(array=Y, pad_width=pad_width, mode="constant")
	else:
		n_pad = 0 

	len_output = 1 + Y.shape[1] - X.shape[1] 

	X = X.astype('float32')
	Y = Y.astype('float32')

	seqlet_neighbors = seqlet_neighbors.astype('int32')
	scores = np.zeros((Y.shape[0], seqlet_neighbors.shape[1], len_output), dtype='float32')
	_jaccard(X, Y, seqlet_neighbors, scores)

	if return_sparse == True:
		return scores.max(axis=-1)

	argmaxs = np.argmax(scores, axis=-1)
	idxs = np.arange(seqlet_neighbors.shape[1])
	results = np.zeros((Y.shape[0], seqlet_neighbors.shape[1], 2))
	for i in range(Y.shape[0]):
		results[i, :, 0] = scores[i][idxs, argmaxs[i]]
		results[i, :, 1] = argmaxs[i] - n_pad

	return results

@njit(parallel=True)
def pairwise_jaccard(X, k):
	n, m = X.shape

	jaccards = np.empty((n, k), dtype='float64')
	neighbors = np.empty((n, k), dtype='int32')

	for i in prange(n):
		jaccard_ = np.empty(n, dtype='float64')

		for j in range(n):
			min_sum = 0.0
			max_sum = 0.0

			for l in range(m):
				sign = np.sign(X[i, l]) * np.sign(X[j, l])
				xi = abs(X[i, l])
				xj = abs(X[j, l])

				if xi > xj:
					min_sum += xj * sign
					max_sum += xi
				else:
					min_sum += xi * sign
					max_sum += xj 

			jaccard_[j] = min_sum / max_sum

		idxs = np.argsort(-jaccard_, kind='mergesort')[:k]

		jaccards[i] = jaccard_[idxs]
		neighbors[i] = idxs

	return jaccards, neighbors


@njit('void(float32[:, :, :], float32[:, :, :], int32[:, :], float32[:, :, :])', parallel=True)
def _jaccard(X, Y, neighbors, scores):
	nx, d, m = X.shape
	ny = Y.shape[0]
	len_output = scores.shape[-1]

	for l in prange(ny):
		for idx in range(len_output):
			for i in range(neighbors.shape[1]):
				min_sum = 0.0
				max_sum = 0.0
				neighbor_li = neighbors[l, i]

				for j in range(idx, idx+d):
					j_idx = j - idx

					for k in range(m):
						sign = np.sign(X[neighbor_li, j_idx, k]) * np.sign(Y[l, j, k])

						x = abs(X[neighbor_li, j_idx, k])
						y = abs(Y[l, j, k])

						if y > x:
							min_sum += x * sign
							max_sum += y
						else:
							min_sum += y * sign
							max_sum += x

				scores[l, i, idx] = min_sum / max_sum



def pearson_correlation(X, Y, min_overlap=None, func=np.ceil):
	if X.ndim == 2:
		X = X[None, :, :]
	if Y.ndim == 2:
		Y = Y[None, :, :]

	if min_overlap is not None:
		n_pad = int(func(X.shape[1]*(1-min_overlap)))
		pad_width = ((0, 0), (n_pad, n_pad), (0, 0)) 
		Y = np.pad(array=Y, pad_width=pad_width, mode="constant")

	n, d, _ = X.shape
	len_output = 1 + Y.shape[1] - d 
	scores = np.zeros((n, len_output))

	for idx in range(len_output):
		Y_ = Y[:, idx:idx+d]

		scores_ = np.dot((X / np.linalg.norm(X)).ravel(),
				  (Y_ / np.linalg.norm(Y_)).ravel()) 
		scores_ = np.nan_to_num(scores_)
		scores[:,idx] = scores_

	argmaxs = np.argmax(scores, axis=1)
	idxs = np.arange(len(scores))
	return np.array([[scores[idxs, argmaxs], argmaxs - n_pad]]).transpose(0, 2, 1)


class NNTsneConditionalProbs():
	def __init__(self, perplexity):
		self.perplexity = perplexity 

	def __call__(self, affinity_mat, nearest_neighbors):
		distmat_nn = np.log((1.0/(0.5*np.maximum(affinity_mat, 0.0000001)))-1)
		distmat_nn = np.maximum(distmat_nn, 0.0) #eliminate tiny neg floats

		# Compute the number of nearest neighbors to find.
		# LvdM uses 3 * perplexity as the number of neighbors.
		# In the event that we have very small # of points
		# set the neighbors to n - 1.
		n_samples = distmat_nn.shape[0]
		k = min(n_samples - 1, int(3. * self.perplexity + 1))

		P = self.tsne_probs_calc(distances_nn=distmat_nn[:,1:(k+1)],
								 neighbors_nn=[row[1:(k+1)] for row in 
											   nearest_neighbors])
		return P

	def tsne_probs_calc(self, distances_nn, neighbors_nn):
		# Compute conditional probabilities such that they approximately match
		# the desired perplexity
		n_samples, k = len(neighbors_nn),len(neighbors_nn[0])
		distances = distances_nn.astype(np.float32, copy=False)
		neighbors = neighbors_nn
		
		conditional_P = sklearn.manifold._utils._binary_search_perplexity(
			distances, self.perplexity, verbose=False)

		eps = 1e-8
		marginal_sum = conditional_P.sum(axis=-1)
		marginal_sum[marginal_sum < eps] = eps

		#normalize the conditional_P to sum to 1 across the rows
		conditional_P = conditional_P / marginal_sum[:,None]

		data = []
		rows = []
		cols = []
		for row_idx,(ps,neigh_row) in enumerate(zip(conditional_P, neighbors)):
			data.extend([p for p,neighbor in zip(ps, neigh_row)])
			rows.extend([row_idx for neighbor in neigh_row])
			cols.extend([neighbor for neighbor in neigh_row])

		P = coo_matrix((data, (rows, cols)),
					   shape=(len(neighbors), len(neighbors)))
		return P




# extract_seqlets.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import numpy as np

from . import core
from sklearn.isotonic import IsotonicRegression

def _bin_mode(values, bins=1000):
	hist, bin_edges = np.histogram(values, bins=bins)
	peak = np.argmax(hist)
	l_edge = bin_edges[peak]
	r_edge = bin_edges[peak+1]
	return l_edge, r_edge, values[(l_edge < values) & (values < r_edge)]

def _laplacian_null(track, window_size, num_to_samp, random_seed=1234):
	percentiles_to_use = np.array([5*(x+1) for x in range(19)])

	rng = np.random.RandomState()
	values = np.concatenate(track, axis=0)

	# first estimate mu, using two level histogram to get to 1e-6
	_, _, top_values = _bin_mode(values)
	l_edge, r_edge, _ = _bin_mode(top_values)
	mu = (l_edge + r_edge) / 2

	pos_values = values[values >= mu]
	neg_values = values[values <=mu] 

	#Take the most aggressive lambda over all percentiles
	pos_laplace_lambda = np.max(
		-np.log(1-(percentiles_to_use/100.0))/
		(np.percentile(a=pos_values, q=percentiles_to_use)-mu))

	neg_laplace_lambda = np.max(
		-np.log(1-(percentiles_to_use/100.0))/
		(np.abs(np.percentile(a=neg_values,
							  q=100-percentiles_to_use)-mu)))

	rng.seed(random_seed)
	prob_pos = float(len(pos_values))/(len(pos_values)+len(neg_values)) 
	sampled_vals = []

	for i in range(num_to_samp):
		sign = 1 if (rng.uniform() < prob_pos) else -1
		if sign == 1:
			sampled_cdf = rng.uniform()
			val = -np.log(1-sampled_cdf)/pos_laplace_lambda + mu 
		else:
			sampled_cdf = rng.uniform() 
			val = mu + np.log(1-sampled_cdf)/neg_laplace_lambda
		sampled_vals.append(val)

	sampled_vals = np.array(sampled_vals)
	return sampled_vals[sampled_vals >= 0], sampled_vals[sampled_vals < 0]


def _iterative_extract_seqlets(score_track, window_size, flank, suppress):
	n, d = score_track.shape
	seqlets = []
	for example_idx, single_score_track in enumerate(score_track):
		while True:
			argmax = np.argmax(single_score_track, axis=0)
			max_val = single_score_track[argmax]

			#bail if exhausted everything that passed the threshold
			#and was not suppressed
			if max_val == -np.inf:
				break

			#need to be able to expand without going off the edge
			if argmax >= flank and argmax < (d-flank): 
				seqlet = core.Seqlet(example_idx=example_idx, 
					start=argmax-flank, end=argmax+window_size+flank,
					is_revcomp=False)

				seqlets.append(seqlet)

			#suppress the chunks within +- suppress
			l_idx = int(max(np.floor(argmax+0.5-suppress),0))
			r_idx = int(min(np.ceil(argmax+0.5+suppress), d))
			single_score_track[l_idx:r_idx] = -np.inf 

	return seqlets


def _smooth_and_split(tracks, window_size, subsample_cap=1000000):
	n = len(tracks)

	tracks = np.hstack([np.zeros((n, 1)), np.cumsum(tracks, axis=-1)])
	tracks = tracks[:, window_size:] - tracks[:, :-window_size]

	values = np.concatenate(tracks, axis=0)
	if len(values) > subsample_cap:
		values = np.random.RandomState(1234).choice(
			a=values, size=subsample_cap, replace=False)
	
	pos_values = values[values >= 0]
	pos_values = np.sort(pos_values)

	neg_values = values[values < 0]
	neg_values = np.sort(neg_values)[::-1]
	return pos_values, neg_values, tracks

def _isotonic_thresholds(values, null_values, increasing, target_fdr, 
	min_frac_neg=0.95):
	n1, n2 = len(values), len(null_values)

	X = np.concatenate([values, null_values], axis=0)
	y = np.concatenate([np.ones(n1), np.zeros(n2)], axis=0)

	w = len(values) / len(null_values)
	sample_weight = np.concatenate([np.ones(n1), np.ones(n2)*w], axis=0)

	model = IsotonicRegression(out_of_bounds='clip', increasing=increasing)
	model.fit(X, y, sample_weight=sample_weight)

	min_prec_x = model.X_min_ if increasing else model.X_max_
	min_precision = model.transform([min_prec_x])[0]
	implied_frac_neg = -1/(1-(1/max(min_precision, 1e-7)))
	if (implied_frac_neg > 1.0 or implied_frac_neg < min_frac_neg):
		implied_frac_neg = max(min(1.0,implied_frac_neg), min_frac_neg)

	precisions = np.minimum(np.maximum(1 + implied_frac_neg*(
		1 - (1 / np.maximum(model.transform(values), 1e-7))), 0.0), 1.0)
	precisions[-1] = 1
	return values[precisions >= (1 - target_fdr)][0]

def _refine_thresholds(vals, pos_threshold, neg_threshold,
	min_passing_windows_frac, max_passing_windows_frac):
	frac_passing_windows =(
		sum(vals >= pos_threshold)
		 + sum(vals <= neg_threshold))/float(len(vals))

	if frac_passing_windows < min_passing_windows_frac:
		pos_threshold = np.percentile(
			a=np.abs(vals),
			q=100*(1-min_passing_windows_frac)) 
		neg_threshold = -pos_threshold

	if frac_passing_windows > max_passing_windows_frac:
		pos_threshold = np.percentile(
			a=np.abs(vals),
			q=100*(1-max_passing_windows_frac)) 
		neg_threshold = -pos_threshold

	return pos_threshold, neg_threshold


def extract_seqlets(attribution_scores, window_size, flank, suppress, 
	target_fdr, min_passing_windows_frac, max_passing_windows_frac, 
	weak_threshold_for_counting_sign):

	pos_values, neg_values, smoothed_tracks = _smooth_and_split(
		attribution_scores, window_size)

	pos_null_values, neg_null_values = _laplacian_null(track=smoothed_tracks, 
		window_size=window_size, num_to_samp=10000)

	pos_threshold = _isotonic_thresholds(pos_values, pos_null_values, 
		increasing=True, target_fdr=target_fdr)
	neg_threshold = _isotonic_thresholds(neg_values, neg_null_values,
		increasing=False, target_fdr=target_fdr)

	pos_threshold, neg_threshold = _refine_thresholds(
		  vals=np.concatenate([pos_values, neg_values], axis=0),
		  pos_threshold=pos_threshold,
		  neg_threshold=neg_threshold,
		  min_passing_windows_frac=min_passing_windows_frac,
		  max_passing_windows_frac=max_passing_windows_frac) 

	distribution = np.array(sorted(np.abs(np.concatenate(smoothed_tracks,
		axis=0))))

	transformed_pos_threshold = np.sign(pos_threshold)*np.searchsorted(
		a=distribution, v=abs(pos_threshold))/len(distribution)

	transformed_neg_threshold = np.sign(neg_threshold)*np.searchsorted(
		a=distribution, v=abs(neg_threshold))/len(distribution)

	idxs = (smoothed_tracks >= pos_threshold) | (smoothed_tracks <= neg_threshold)

	smoothed_tracks[idxs] = np.abs(smoothed_tracks[idxs])
	smoothed_tracks[~idxs] = -np.inf

	# Filter out the flanks
	smoothed_tracks[:, :flank] = -np.inf
	smoothed_tracks[:, -flank:] = -np.inf

	seqlets = _iterative_extract_seqlets(score_track=smoothed_tracks,
		window_size=window_size,
		flank=flank,
		suppress=suppress)

	#find the weakest transformed threshold used across all tasks
	weak_thresh = min(min(transformed_pos_threshold, 
		abs(transformed_neg_threshold)) - 0.0001, 
			weak_threshold_for_counting_sign)

	threshold = distribution[int(weak_thresh * len(distribution))]

	return seqlets, threshold



from . import aggregator
from . import affinitymat
from . import extract_seqlets
from . import core
from . import gapped_kmer
from . import io
from . import util
from . import tfmodisco
from . import cluster
from . import report

__version__ = '2.0.7'


# util.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import numpy as np
from numba import njit

def cpu_sliding_window_sum(arr, window_size):
	to_return = np.zeros(len(arr)-window_size+1)
	current_sum = np.sum(arr[0:window_size])
	to_return[0] = current_sum
	idx_to_include = window_size
	idx_to_exclude = 0
	while idx_to_include < len(arr):
		current_sum += (arr[idx_to_include] - arr[idx_to_exclude]) 
		to_return[idx_to_exclude+1] = current_sum
		idx_to_include += 1
		idx_to_exclude += 1
	return to_return


@njit('float64(float64, float64[:])')
def binary_search_perplexity(desired_perplexity, distances):
	EPSILON_DBL = 1e-8
	PERPLEXITY_TOLERANCE = 1e-5
	n_steps = 100
	
	desired_entropy = np.log(desired_perplexity)
	
	beta_min = -np.inf
	beta_max = np.inf
	beta = 1.0
	
	for l in range(n_steps):
		ps = np.exp(-distances * beta)
		sum_ps = np.sum(ps) + 1
		ps = ps/(max(sum_ps,EPSILON_DBL))
		sum_disti_Pi = np.sum(distances*ps)
		entropy = np.log(sum_ps) + beta * sum_disti_Pi
		
		entropy_diff = entropy - desired_entropy
		if np.abs(entropy_diff) <= PERPLEXITY_TOLERANCE:
			break
		
		if entropy_diff > 0.0:
			beta_min = beta
			if beta_max == np.inf:
				beta *= 2.0
			else:
				beta = (beta + beta_max) / 2.0
		else:
			beta_max = beta
			if beta_min == -np.inf:
				beta /= 2.0
			else:
				beta = (beta + beta_min) / 2.0
	
	return beta


def compute_per_position_ic(ppm, background, pseudocount):
	"""Compute information content at each position of ppm.

	Arguments:
		ppm: should have dimensions of length x alphabet. Entries along the
			alphabet axis should sum to 1.
		background: the background base frequencies
		pseudocount: pseudocount to be added to the probabilities of the ppm
			to prevent overflow/underflow.

	Returns:
		total information content at each positon of the ppm.
	"""

	if (not np.allclose(np.sum(ppm, axis=1), 1.0, atol=1.0e-5)):
		ppm = ppm/np.sum(ppm, axis=1)[:,None]

	alphabet_len = len(background)
	ic = ((np.log((ppm+pseudocount)/(1 + pseudocount*alphabet_len))/np.log(2))
		  *ppm - (np.log(background)*background/np.log(2))[None,:])
	return np.sum(ic,axis=1)


#rolling_window is from this blog post by Erik Rigtorp:
# https://rigtorp.se/2011/01/01/rolling-statistics-numpy.html
#The last axis of a will be subject to the windowing
def rolling_window(a, window):
	shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
	strides = a.strides + (a.strides[-1],)
	return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def magnitude(X):
	X = X - np.mean(X)
	return (X / (np.linalg.norm(X.ravel())+0.0000001))

def l1(X):
	abs_sum = np.sum(np.abs(X))
	if abs_sum == 0:
		return X
	return (X/abs_sum)

def get_2d_data_from_patterns(patterns, transformer='l1', include_hypothetical=True):
	func = l1 if transformer == 'l1' else magnitude
	tracks = ['hypothetical_contribs', 'contrib_scores']
	if not include_hypothetical:
		tracks = tracks[1:]

	all_fwd_data, all_rev_data = [], []

	for pattern in patterns:
		snippets = [getattr(pattern, track) for track in tracks]

		fwd_data = np.concatenate([func(snippet) for snippet in snippets], axis=1)
		rev_data = np.concatenate([func(snippet[::-1, ::-1]) for snippet in snippets], axis=1)

		all_fwd_data.append(fwd_data)
		all_rev_data.append(rev_data)

	return np.array(all_fwd_data), np.array(all_rev_data)



# io.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>

import h5py
import hdf5plugin

import numpy as np

def convert(old_filename, filename):
	old_grp = h5py.File(old_filename, "r")['metacluster_idx_to_submetacluster_results']
	new_grp = h5py.File(filename, "w")

	if 'metacluster_0' in old_grp.keys():
		pos_patterns_grp = new_grp.create_group("pos_patterns")
		old_patterns_grp = old_grp['metacluster_0']['seqlets_to_patterns_result']

		if 'patterns' in old_patterns_grp:
			old_patterns_grp = old_patterns_grp['patterns'] 

			for pattern in old_patterns_grp['all_pattern_names'][:]:
				old_pattern = old_patterns_grp[pattern]

				sequence = old_pattern['sequence']['fwd'][:]
				contrib_scores = old_pattern['task0_contrib_scores']['fwd'][:]
				hypothetical_contribs = old_pattern['task0_hypothetical_contribs']['fwd'][:]


				pattern_grp = pos_patterns_grp.create_group(pattern)
				pattern_grp.create_dataset("sequence", data=sequence)
				pattern_grp.create_dataset("contrib_scores", data=contrib_scores)
				pattern_grp.create_dataset("hypothetical_contribs", data=hypothetical_contribs)
				
				seqlet_grp = pattern_grp.create_group("seqlets")
				seqlet_grp.create_dataset("n_seqlets", 
					data=np.array([len(old_pattern['seqlets_and_alnmts']['seqlets'])]))

				if 'subcluster_to_subpattern' in old_pattern.keys():
					old_subpatterns_grp = old_pattern['subcluster_to_subpattern']
					for subpattern in old_subpatterns_grp['subcluster_names'][:]:
						old_subpattern = old_subpatterns_grp[subpattern]

						sequence = old_subpattern['sequence']['fwd'][:]
						contrib_scores = old_subpattern['task0_contrib_scores']['fwd'][:]
						hypothetical_contribs = old_subpattern['task0_hypothetical_contribs']['fwd'][:]

						subpattern_grp = pattern_grp.create_group(subpattern)
						subpattern_grp.create_dataset("sequence", data=sequence)
						subpattern_grp.create_dataset("contrib_scores", data=contrib_scores)
						subpattern_grp.create_dataset("hypothetical_contribs", data=hypothetical_contribs)
						
						seqlet_grp = subpattern_grp.create_group("seqlets")
						seqlet_grp.create_dataset("n_seqlets", 
							data=np.array([len(old_subpattern['seqlets_and_alnmts']['seqlets'])]))

	if 'metacluster_1' in old_grp.keys():
		neg_patterns_grp = new_grp.create_group("neg_patterns")
		old_patterns_grp = old_grp['metacluster_1']['seqlets_to_patterns_result']

		if 'patterns' in old_patterns_grp:
			old_patterns_grp = old_patterns_grp['patterns']

			for pattern in old_patterns_grp['all_pattern_names'][:]:
				old_pattern = old_patterns_grp[pattern]

				sequence = old_pattern['sequence']['fwd'][:]
				contrib_scores = old_pattern['task0_contrib_scores']['fwd'][:]
				hypothetical_contribs = old_pattern['task0_hypothetical_contribs']['fwd'][:]

				pattern_grp = neg_patterns_grp.create_group(pattern)
				pattern_grp.create_dataset("sequence", data=sequence)
				pattern_grp.create_dataset("contrib_scores", data=contrib_scores)
				pattern_grp.create_dataset("hypothetical_contribs", data=hypothetical_contribs)

				seqlet_grp = pattern_grp.create_group("seqlets")
				seqlet_grp.create_dataset("n_seqlets", 
					data=np.array([len(old_pattern['seqlets_and_alnmts']['seqlets'])]))

				if 'subcluster_to_subpattern' in old_pattern.keys():
					old_subpatterns_grp = old_pattern['subcluster_to_subpattern']
					for subpattern in old_subpatterns_grp['subcluster_names'][:]:
						old_subpattern = old_subpatterns_grp[subpattern]
						sequence = old_subpattern['sequence']['fwd'][:]
						contrib_scores = old_subpattern['task0_contrib_scores']['fwd'][:]
						hypothetical_contribs = old_subpattern['task0_hypothetical_contribs']['fwd'][:]

						subpattern_grp = pattern_grp.create_group(subpattern)
						subpattern_grp.create_dataset("sequence", data=sequence)
						subpattern_grp.create_dataset("contrib_scores", data=contrib_scores)
						subpattern_grp.create_dataset("hypothetical_contribs", data=hypothetical_contribs)

						seqlet_grp = subpattern_grp.create_group("seqlets")
						seqlet_grp.create_dataset("n_seqlets", 
							data=np.array([len(old_subpattern['seqlets_and_alnmts']['seqlets'])]))


def save_pattern(pattern, grp):
	grp.create_dataset("sequence", data=pattern.sequence)
	grp.create_dataset("contrib_scores", data=pattern.contrib_scores)
	grp.create_dataset("hypothetical_contribs", data=pattern.hypothetical_contribs)

	seqlet_grp = grp.create_group("seqlets")
	seqlet_grp.create_dataset("n_seqlets", data=np.array([len(pattern.seqlets)]))
	seqlet_grp.create_dataset("start", 
		data=np.array([seqlet.start for seqlet in pattern.seqlets]))
	seqlet_grp.create_dataset("end", 
		data=np.array([seqlet.end for seqlet in pattern.seqlets]))
	seqlet_grp.create_dataset("example_idx", 
		data=np.array([seqlet.example_idx for seqlet in pattern.seqlets]))
	seqlet_grp.create_dataset("is_revcomp",
		data=np.array([seqlet.is_revcomp for seqlet in pattern.seqlets]))

	seqlet_grp.create_dataset("sequence",
		data=np.array([seqlet.sequence for seqlet in pattern.seqlets]))
	seqlet_grp.create_dataset("contrib_scores",
		data=np.array([seqlet.contrib_scores for seqlet in pattern.seqlets]))
	seqlet_grp.create_dataset("hypothetical_contribs",
		data=np.array([seqlet.hypothetical_contribs for seqlet in pattern.seqlets]))

	if pattern.subclusters is not None:
		for subcluster, subpattern in pattern.subcluster_to_subpattern.items():
			subpattern_grp = grp.create_group("subpattern_"+str(subcluster)) 
			save_pattern(subpattern, subpattern_grp)


def save_hdf5(filename, pos_patterns, neg_patterns):
	"""Save the results of tf-modisco to a h5 file.

	This function will save the SeqletSets and their associated seqlets in
	a minimal new minimal format or in the older format originally used by
	TF-MoDISco. Regardless of format, only information used in the SeqletSets
	are saved.


	Parameters
	----------
	filename: str
		The name of the h5 file to save to.

	pos_patterns: list or None
		A list of SeqletSet objects or None.

	neg_patterns: list or None
		A list of SeqletSet objects or None.
	"""

	grp = h5py.File(filename, 'w')
	
	if pos_patterns is not None:
		pos_group = grp.create_group("pos_patterns")
		for idx, pattern in enumerate(pos_patterns):
			pos_pattern = pos_group.create_group("pattern_"+str(idx))
			save_pattern(pattern, pos_pattern)

	if neg_patterns is not None:
		neg_group = grp.create_group("neg_patterns")
		for idx, pattern in enumerate(neg_patterns):
			neg_pattern = neg_group.create_group("pattern_"+str(idx))
			save_pattern(pattern, neg_pattern)


def convert_new_to_old(new_format_filename, old_format_filename):
	'''This function does the opposite of the convert function.
	Given the filepath to a tfmodisco-lite-formatted file (arg 1),
	it writes the same information in the original tfmodisco format
	to file (arg 2).
	
	This function assumes that metacluster0 should be the positive
	patterns (patterns formed from positive importance scores, and 
	that metacluster1 is the negative patterns. It also assumes that
	there is only 1 task (the standard use-case).
	
	This function does not fill in all of the information for the
	original modisco format, in part because some of that information
	is no longer in the new modisco format. But the info converted is
	sufficient to run motif hit calling using the original modisco algorithm.
	'''

	old_f = h5py.File(old_format_filename, "w")
	new_f = h5py.File(new_format_filename, "r")

	old_f.create_dataset("task_names", data=["task0"])

	old_fmt_grp = old_f.create_group('metacluster_idx_to_submetacluster_results')
    
	patterns_group_name_to_metacluster_name = {"pos_patterns" : 'metacluster_0', "neg_patterns" : 'metacluster_1'}
    
	for patterns_group_name in ['pos_patterns', 'neg_patterns']:
		if patterns_group_name in new_f.keys():
			metacluster_name = patterns_group_name_to_metacluster_name[patterns_group_name]
			metacluster_seqlet_strings = []
			
			# new format
			new_patterns_grp = new_f[patterns_group_name]
			
			# old format
			old_metacluster_grp = old_fmt_grp.create_group(metacluster_name)
			old_patterns_grp = old_metacluster_grp.create_group('seqlets_to_patterns_result')
			
			# these needed to avoid error / silent failure
			old_patterns_grp.attrs["success"] = True
			old_patterns_grp.attrs["total_time_taken"] = 1.0
			
			# if there are any patterns for this hit (should always be???)...
			if len(new_patterns_grp.keys()) > 0:
				pattern_names = list(new_patterns_grp.keys())
				# needed because otherwise order is (0, 1, 11, 12, ...)
				pattern_names = sorted(pattern_names, key = lambda name : int(name.split("_")[1]))
				
				old_patterns_subgrp = old_patterns_grp.create_group("patterns")
				
				# for each modisco hit...
				for pattern in pattern_names:
					pattern_grp = new_patterns_grp[pattern]
					
					# new format
					sequence = pattern_grp["sequence"]
					contrib_scores = pattern_grp["contrib_scores"]
					hypothetical_contribs = pattern_grp["hypothetical_contribs"]
					
					# old format
					old_pattern_grp = old_patterns_subgrp.create_group(pattern)
					old_pattern_grp.create_dataset("sequence/fwd", data=sequence)
					old_pattern_grp.create_dataset("task0_contrib_scores/fwd", data=contrib_scores)
					old_pattern_grp.create_dataset("task0_hypothetical_contribs/fwd", data=hypothetical_contribs)
					
					seqlets_grp = pattern_grp['seqlets']
					
					# in the old format, seqlets were stored as a list of strings
					seqlet_strings = []
					for i in range(len(seqlets_grp['example_idx'])):
						# new format: separate arrays for each seqlet attribute,
						# where each array is len(# seqlets)
						example_idx = str(seqlets_grp['example_idx'][i])
						start = str(seqlets_grp['start'][i])
						end = str(seqlets_grp['end'][i])
						rc = str(seqlets_grp['is_revcomp'][i])
						seqlet_str = "example:" + example_idx + ",start:" + start + ",end:" + end + ",rc:" + rc
						seqlet_strings.append(seqlet_str)
					metacluster_seqlet_strings.extend(seqlet_strings)
					
					# old format
					old_seq_align_grp = old_pattern_grp.create_group("seqlets_and_alnmts")
					old_seq_align_grp.create_dataset('seqlets', data=seqlet_strings)
					
					# dummy data to avoid error: alignments are all 0s
					# (must be len(# seqlets), and must be ints)
					old_seq_align_grp.create_dataset('alnmts', data=np.zeros((len(seqlet_strings),)), dtype="i")
					
					# repeat the process for each pattern/cluster (above) for each subpattern/subcluster (below)
					
					subcluster_names = [k for k in pattern_grp.keys() if k.startswith("subcluster_")]
					if len(subcluster_names) > 0:
						old_subpatterns_grp = old_pattern_grp.create_group("subcluster_to_subpattern")
						
						for subpattern in subcluster_names:
							subpattern_grp = pattern_grp[subpattern]
							sequence = subpattern_grp["sequence"]
							contrib_scores = subpattern_grp["contrib_scores"]
							hypothetical_contribs = subpattern_grp["hypothetical_contribs"]

							old_subpattern_grp = old_subpatterns_grp.create_group(subpattern)
							old_subpattern_grp.create_dataset("sequence/fwd", data=sequence)
							old_subpattern_grp.create_dataset("task0_contrib_scores/fwd", data=contrib_scores)
							old_subpattern_grp.create_dataset("task0_hypothetical_contribs/fwd", data=hypothetical_contribs)

							seqlets_grp = subpattern_grp['seqlets']

							seqlet_strings = []
							for i in range(len(seqlets_grp['example_idx'])):
								example_idx = str(seqlets_grp['example_idx'][i])
								start = str(seqlets_grp['start'][i])
								end = str(seqlets_grp['end'][i])
								rc = str(seqlets_grp['is_revcomp'][i])
								seqlet_str = "example:" + example_idx + ",start:" + start + ",end:" + end + ",rc:" + rc
								seqlet_strings.append(seqlet_str)

							old_seq_align_grp = old_subpattern_grp.create_group("seqlets_and_alnmts")
							old_seq_align_grp.create_dataset('seqlets', data=seqlet_strings)
							old_seq_align_grp.create_dataset('alnmts', data=np.zeros((len(seqlet_strings),)), dtype="i")

						old_subpatterns_grp.create_dataset("subcluster_names", data=subcluster_names)
                    
				old_patterns_subgrp.create_dataset("all_pattern_names", data=pattern_names)

			# required to avoid error: a collection of seqlets for the entire metacluster
			old_metacluster_grp.create_dataset("seqlets", data=metacluster_seqlet_strings)

	old_f.close()
	new_f.close()


import os
import pickle
import h5py
import pandas
import tempfile

import matplotlib
matplotlib.use('pdf')
from matplotlib import pyplot as plt

import pandas as pd
import numpy as np

import logomaker

pd.options.display.max_colwidth = 500

def read_meme(filename):
	motifs = {}

	with open(filename, "r") as infile:
		motif, width, i = None, None, 0

		for line in infile:
			if motif is None:
				if line[:5] == 'MOTIF':
					motif = line.split()[1]
				else:
					continue

			elif width is None:
				if line[:6] == 'letter':
					width = int(line.split()[5])
					pwm = np.zeros((width, 4))

			elif i < width:
				pwm[i] = list(map(float, line.split()))
				i += 1

			else:
				motifs[motif] = pwm
				motif, width, i = None, None, 0

	return motifs


def compute_per_position_ic(ppm, background, pseudocount):
    alphabet_len = len(background)
    ic = ((np.log((ppm+pseudocount)/(1 + pseudocount*alphabet_len))/np.log(2))
          *ppm - (np.log(background)*background/np.log(2))[None,:])
    return np.sum(ic,axis=1)


def write_meme_file(ppm, bg, fname):
	f = open(fname, 'w')
	f.write('MEME version 4\n\n')
	f.write('ALPHABET= ACGT\n\n')
	f.write('strands: + -\n\n')
	f.write('Background letter frequencies (from unknown source):\n')
	f.write('A %.3f C %.3f G %.3f T %.3f\n\n' % tuple(list(bg)))
	f.write('MOTIF 1 TEMP\n\n')
	f.write('letter-probability matrix: alength= 4 w= %d nsites= 1 E= 0e+0\n' % ppm.shape[0])
	for s in ppm:
		f.write('%.5f %.5f %.5f %.5f\n' % tuple(s))
	f.close()


def fetch_tomtom_matches(ppm, cwm, motifs_db, 
	background=[0.25, 0.25, 0.25, 0.25], tomtom_exec_path='tomtom',
	trim_threshold=0.3, trim_min_length=3):

	"""Fetches top matches from a motifs database using TomTom.
	Args:
		ppm: position probability matrix- numpy matrix of dimension (N,4)
		background: list with ACGT background probabilities
		tomtom_exec_path: path to TomTom executable
		motifs_db: path to motifs database in meme format
		n: number of top matches to return, ordered by p-value
		temp_dir: directory for storing temp files
		trim_threshold: the ppm is trimmed from left till first position for which
			probability for any base pair >= trim_threshold. Similarly from right.
	Returns:
		list: a list of up to n results returned by tomtom, each entry is a
			dictionary with keys 'Target ID', 'p-value', 'E-value', 'q-value'
	"""

	_, fname = tempfile.mkstemp()
	_, tomtom_fname = tempfile.mkstemp()

	score = np.sum(np.abs(cwm), axis=1)
	trim_thresh = np.max(score) * trim_threshold  # Cut off anything less than 30% of max score
	pass_inds = np.where(score >= trim_thresh)[0]
	trimmed = ppm[np.min(pass_inds): np.max(pass_inds) + 1]

	# can be None of no base has prob>t
	if trimmed is None:
		return []

	# trim and prepare meme file
	write_meme_file(trimmed, background, fname)

	# run tomtom
	cmd = '%s -no-ssc -oc . --verbosity 1 -text -min-overlap 5 -mi 1 -dist pearson -evalue -thresh 10.0 %s %s > %s' % (tomtom_exec_path, fname, motifs_db, tomtom_fname)

	os.system(cmd)
	tomtom_results = pandas.read_csv(tomtom_fname, sep="\t", usecols=(1, 5))
	os.system('rm ' + tomtom_fname)
	os.system('rm ' + fname)
	return tomtom_results


def run_tomtom(modisco_h5py, output_prefix, meme_motif_db, top_n_matches=3, 
	tomtom_exec="tomtom", trim_threshold=0.3, trim_min_length=3):
	modisco_results = h5py.File(modisco_h5py, 'r')

	tomtom_results = {'pattern': [], 'num_seqlets': []}
	for i in range(top_n_matches):
		tomtom_results['match{}'.format(i)] = []
		tomtom_results['qval{}'.format(i)] = []

	for name in ['pos_patterns', 'neg_patterns']:
		if name not in modisco_results.keys():
			continue

		metacluster = modisco_results[name]
		key = lambda x: int(x[0].split("_")[-1])
		for pattern_name, pattern in sorted(metacluster.items(), key=key):
			ppm = np.array(pattern['sequence'][:])
			cwm = np.array(pattern["contrib_scores"][:])

			num_seqlets = pattern['seqlets']['n_seqlets'][:][0]
			tag = '{}.{}'.format(name, pattern_name)

			r = fetch_tomtom_matches(ppm, cwm, motifs_db=meme_motif_db,
				tomtom_exec_path=tomtom_exec, trim_threshold=trim_threshold,
				trim_min_length=trim_min_length)

			tomtom_results['pattern'].append(tag)
			tomtom_results['num_seqlets'].append(num_seqlets)

			i = -1
			for i, (target, qval) in r.iloc[:top_n_matches].iterrows():
				tomtom_results['match{}'.format(i)].append(target)
				tomtom_results['qval{}'.format(i)].append(qval)

			for j in range(i+1, top_n_matches):
				tomtom_results['match{}'.format(j)].append(None)
				tomtom_results['qval{}'.format(j)].append(None)			

	modisco_results.close()
	return pandas.DataFrame(tomtom_results)


def path_to_image_html(path):
	return '<img src="'+ path + '" width="240" >'

def _plot_weights(array, path, figsize=(10,3), **kwargs):
	fig = plt.figure(figsize=figsize)
	ax = fig.add_subplot(111) 

	df = pandas.DataFrame(array, columns=['A', 'C', 'G', 'T'])
	df.index.name = 'pos'

	crp_logo = logomaker.Logo(df, ax=ax)
	crp_logo.style_spines(visible=False)
	plt.ylim(min(df.sum(axis=1).min(), 0), df.sum(axis=1).max())

	plt.savefig(path)
	plt.close()
	
def make_logo(match, logo_dir, motifs):
	if match == 'NA':
		return

	background = np.array([0.25, 0.25, 0.25, 0.25])
	ppm = motifs[match]
	ic = compute_per_position_ic(ppm, background, 0.001)

	_plot_weights(ppm*ic[:, None], path='{}/{}.png'.format(logo_dir, match))
		

def create_modisco_logos(modisco_file, modisco_logo_dir, trim_threshold):
	results = h5py.File(modisco_file, 'r')
	tags = []

	for name in ["pos_patterns", "neg_patterns"]:
		if name not in results.keys():
			continue

		metacluster = results[name]
		key = lambda x: int(x[0].split("_")[-1])
		for pattern_name, pattern in sorted(metacluster.items(), key=key):
			tag = '{}.{}'.format(name, pattern_name)
			tags.append(tag)

			cwm_fwd = np.array(pattern['contrib_scores'][:])
			cwm_rev = cwm_fwd[::-1, ::-1]

			score_fwd = np.sum(np.abs(cwm_fwd), axis=1)
			score_rev = np.sum(np.abs(cwm_rev), axis=1)

			trim_thresh_fwd = np.max(score_fwd) * trim_threshold
			trim_thresh_rev = np.max(score_rev) * trim_threshold

			pass_inds_fwd = np.where(score_fwd >= trim_thresh_fwd)[0]
			pass_inds_rev = np.where(score_rev >= trim_thresh_rev)[0]

			start_fwd, end_fwd = max(np.min(pass_inds_fwd) - 4, 0), min(np.max(pass_inds_fwd) + 4 + 1, len(score_fwd) + 1)
			start_rev, end_rev = max(np.min(pass_inds_rev) - 4, 0), min(np.max(pass_inds_rev) + 4 + 1, len(score_rev) + 1)

			trimmed_cwm_fwd = cwm_fwd[start_fwd:end_fwd]
			trimmed_cwm_rev = cwm_rev[start_rev:end_rev]

			_plot_weights(trimmed_cwm_fwd, path='{}/{}.cwm.fwd.png'.format(modisco_logo_dir, tag))
			_plot_weights(trimmed_cwm_rev, path='{}/{}.cwm.rev.png'.format(modisco_logo_dir, tag))

	return tags

def report_motifs(modisco_h5py, output_dir, meme_motif_db, suffix='./', 
	top_n_matches=3, trim_threshold=0.3, trim_min_length=3):

	if not os.path.isdir(output_dir):
		os.mkdir(output_dir)

	if not os.path.isdir(output_dir + '/trimmed_logos/'):
		os.mkdir(output_dir + '/trimmed_logos/')
	modisco_logo_dir = output_dir + '/trimmed_logos/'

	motifs = read_meme(meme_motif_db)
	names = create_modisco_logos(modisco_h5py, modisco_logo_dir, trim_threshold)

	tomtom_df = run_tomtom(modisco_h5py, output_dir, meme_motif_db, 
		top_n_matches=top_n_matches, tomtom_exec="tomtom", 
		trim_threshold=trim_threshold, trim_min_length=trim_min_length)

	tomtom_df['modisco_cwm_fwd'] = ['{}trimmed_logos/{}.cwm.fwd.png'.format(suffix, name) for name in names]
	tomtom_df['modisco_cwm_rev'] = ['{}trimmed_logos/{}.cwm.rev.png'.format(suffix, name) for name in names]

	reordered_columns = ['pattern', 'num_seqlets', 'modisco_cwm_fwd', 'modisco_cwm_rev']
	for i in range(top_n_matches):
		name = "match{}".format(i)
		logos = []

		for index, row in tomtom_df.iterrows():
			if name in tomtom_df.columns:
				if pandas.isnull(row[name]):
					logos.append("NA")
				else:
					make_logo(row[name], output_dir, motifs)
					logos.append("{}{}.png".format(suffix, row[name]))
			else:
				break

		tomtom_df["{}_logo".format(name)] = logos
		reordered_columns.extend([name, 'qval{}'.format(i), "{}_logo".format(name)])

	tomtom_df = tomtom_df[reordered_columns]
	tomtom_df.to_html(open('{}/motifs.html'.format(output_dir), 'w'),
		escape=False, formatters=dict(modisco_cwm_fwd=path_to_image_html,
			modisco_cwm_rev=path_to_image_html, match0_logo=path_to_image_html,
			match1_logo=path_to_image_html, match2_logo=path_to_image_html), 
		index=False)



# core.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import pickle

import numpy as np
import scipy.sparse

from . import affinitymat
from . import cluster
from . import util

from collections import OrderedDict

class TrackSet(object):
	def __init__(self, one_hot, contrib_scores, hypothetical_contribs):
		self.one_hot = one_hot
		self.contrib_scores = contrib_scores
		self.hypothetical_contribs = hypothetical_contribs
		self.length = len(one_hot[0])

	def create_seqlets(self, seqlets):
		for seqlet in seqlets:
			idx = seqlet.example_idx
			s, e = seqlet.start, seqlet.end

			if seqlet.is_revcomp:
				seqlet.sequence = self.one_hot[idx][s:e][::-1, ::-1]
				seqlet.contrib_scores = self.contrib_scores[idx][s:e][::-1, ::-1]
				seqlet.hypothetical_contribs = self.hypothetical_contribs[idx][s:e][::-1, ::-1]
			else:
				seqlet.sequence = self.one_hot[idx][s:e]
				seqlet.contrib_scores = self.contrib_scores[idx][s:e]
				seqlet.hypothetical_contribs = self.hypothetical_contribs[idx][s:e]				

		return seqlets

class Seqlet(object):
	def __init__(self, example_idx, start, end, is_revcomp):
		self.example_idx = example_idx
		self.start = start
		self.end = end
		self.is_revcomp = is_revcomp

		self.sequence = None
		self.contrib_scores = None
		self.hypothetical_contribs = None

		super(Seqlet, self).__init__()

	def __str__(self):
		return ("example:"+str(self.example_idx)
				+",start:"+str(self.start)+",end:"+str(self.end)
				+",rc:"+str(self.is_revcomp))

	def __len__(self):
		return self.end - self.start

	@property
	def string(self):
		return str(self.example_idx)+"_"+str(self.start)+"_"+str(self.end)

	def revcomp(self):
		new_seqlet = Seqlet(
				example_idx=self.example_idx,
				start=self.start, end=self.end,
				is_revcomp=(self.is_revcomp==False))

		new_seqlet.sequence = self.sequence[::-1, ::-1]
		new_seqlet.contrib_scores = self.contrib_scores[::-1, ::-1]
		new_seqlet.hypothetical_contribs = self.hypothetical_contribs[::-1, ::-1]
		return new_seqlet

	def shift(self, shift_amt):
		return Seqlet(
				example_idx=self.example_idx,
				start=self.start+shift_amt, end=self.end+shift_amt,
				is_revcomp=self.is_revcomp)

	def trim(self, start_idx, end_idx):
		if self.is_revcomp == False:
			new_start = self.start + start_idx 
			new_end = self.start + end_idx
		else:
			new_start = self.end - end_idx
			new_end = self.end - start_idx

		new_seqlet = Seqlet(example_idx=self.example_idx,
			start=new_start, end=new_end, is_revcomp=self.is_revcomp)

		s, e = start_idx, end_idx
		new_seqlet.sequence = self.sequence[s:e]
		new_seqlet.contrib_scores = self.contrib_scores[s:e]
		new_seqlet.hypothetical_contribs = self.hypothetical_contribs[s:e]
		return new_seqlet


class SeqletSet():
	def __init__(self, seqlets):
		self.seqlets = []
		self.unique_seqlets = {}
		self.length = max([len(seqlet) for seqlet in seqlets])  
		
		self._sequence_sum = np.zeros((self.length, 4), dtype='float')
		self._contrib_sum = np.zeros((self.length, 4), dtype='float')
		self._hypothetical_sum = np.zeros((self.length, 4), dtype='float')

		self.sequence = np.zeros((self.length, 4), dtype='float')
		self.contrib_scores = np.zeros((self.length, 4), dtype='float')
		self.hypothetical_contribs = np.zeros((self.length, 4), dtype='float')

		self.per_position_counts = np.zeros((self.length,))

		for seqlet in seqlets:
			if seqlet.string not in self.unique_seqlets: 
				self._add_seqlet(seqlet=seqlet)
		
		self.subclusters = None
		self.subcluster_to_subpattern = None

	def compute_subpatterns(self, perplexity, n_seeds, n_iterations=-1):
		#this method assumes all the seqlets have been expanded so they
		# all start at 0
		X = util.get_2d_data_from_patterns(self.seqlets)[0]
		X = X.reshape(len(X), -1)
	
		n = len(X)
		n_neighb = min(int(perplexity*3 + 2), len(X))

		affmat_nn, seqlet_neighbors = affinitymat.pairwise_jaccard(X, n_neighb)

		distmat_nn = np.log((1.0/(0.5*np.maximum(affmat_nn, 0.0000001)))-1)
		distmat_nn = np.maximum(distmat_nn, 0.0) #eliminate tiny neg floats

		distmat_sp = scipy.sparse.coo_matrix(
				(np.concatenate(distmat_nn, axis=0),
				 (np.array([i for i in range(len(seqlet_neighbors))
							   for j in seqlet_neighbors[i]]).astype("int"),
				  np.concatenate(seqlet_neighbors, axis=0)) ),
				shape=(len(distmat_nn), len(distmat_nn))).tocsr()

		distmat_sp.sort_indices()

		#do density adaptation
		sp_density_adapted_affmat = affinitymat.NNTsneConditionalProbs(
				perplexity=perplexity)(affmat_nn, seqlet_neighbors)

		sp_density_adapted_affmat += sp_density_adapted_affmat.T
		sp_density_adapted_affmat /= np.sum(sp_density_adapted_affmat.data)

		#Do Leiden clustering
		self.subclusters = cluster.LeidenCluster(sp_density_adapted_affmat,
			n_seeds=n_seeds, n_leiden_iterations=n_iterations) 

		#this method assumes all the seqlets have been expanded so they
		# all start at 0
		subcluster_to_seqletsandalignments = OrderedDict()
		for seqlet, subcluster in zip(self.seqlets, self.subclusters):
			if (subcluster not in subcluster_to_seqletsandalignments):
				subcluster_to_seqletsandalignments[subcluster] = []
			
			subcluster_to_seqletsandalignments[subcluster].append(seqlet)

		subcluster_to_subpattern = OrderedDict([
			(subcluster, SeqletSet(seqletsandalignments))
			for subcluster,seqletsandalignments in
			subcluster_to_seqletsandalignments.items()])

		#resort subcluster_to_subpattern so that the subclusters with the
		# most seqlets come first
		self.subcluster_to_subpattern = OrderedDict(
			sorted(subcluster_to_subpattern.items(),
				   key=lambda x: -len(x[1].seqlets)))

	def copy(self):
		return SeqletSet(seqlets=[seqlet for seqlet in self.seqlets])

	def trim_to_support(self, min_frac, min_num):
		max_support = max(self.per_position_counts)
		num = min(min_num, max_support*min_frac)
		
		left_idx = 0
		while self.per_position_counts[left_idx] < num:
			left_idx += 1

		right_idx = len(self.per_position_counts)
		while self.per_position_counts[right_idx-1] < num:
			right_idx -= 1
		
		return self.trim_to_idx(start_idx=left_idx, end_idx=right_idx) 

	def trim_to_idx(self, start_idx, end_idx):
		new_seqlets = []
		for seqlet in self.seqlets:
				new_seqlet = seqlet.trim(start_idx=start_idx, end_idx=end_idx)
				new_seqlets.append(new_seqlet)
		return SeqletSet(seqlets=new_seqlets)

	def _add_seqlet(self, seqlet):
		n = len(seqlet)

		self.seqlets.append(seqlet)
		self.unique_seqlets[seqlet.string] = seqlet
		self.per_position_counts[:n] += 1.0 

		ppc = self.per_position_counts[:, None]
		ppc = ppc + 1E-7 * (ppc == 0)

		self._sequence_sum[:n] += seqlet.sequence
		self._contrib_sum[:n] += seqlet.contrib_scores
		self._hypothetical_sum[:n] += seqlet.hypothetical_contribs

		self.sequence = self._sequence_sum / ppc
		self.contrib_scores = self._contrib_sum / ppc
		self.hypothetical_contribs = self._hypothetical_sum / ppc

	def __len__(self):
		return self.length

	def save_seqlets(self, filename):
		bases = np.array(['A', 'C', 'G', 'T'])

		with open(filename, "w") as outfile:
			for seqlet in self.seqlets:
				sequence = "".join(bases[np.argmax(seqlet.sequence, axis=-1)])
				example_index = seqlet.example_idx
				start, end = seqlet.start, seqlet.end
				outfile.write(">example%d:%d-%d\n" % (example_index, start, end))
				outfile.write(sequence + "\n")



# aggregator.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import numpy as np

from . import affinitymat
from . import core
from . import util

from collections import OrderedDict
from sklearn.metrics import roc_auc_score


def polish_pattern(pattern, min_frac, min_num, track_set, flank, window_size, bg_freq):
	pattern = pattern.trim_to_support(min_frac=min_frac, min_num=min_num)

	pattern = _expand_seqlets_to_fill_pattern(pattern, 
		track_set=track_set, left_flank_to_add=flank,
		right_flank_to_add=flank)

	if pattern is None:
		return None

	# Trim by IC
	ppm = pattern.sequence
	per_pos_ic = util.compute_per_position_ic(
		ppm=ppm, background=bg_freq, pseudocount=0.001)

	start_idx = np.argmax(util.cpu_sliding_window_sum(
		arr=per_pos_ic, window_size=window_size))

	end_idx = start_idx + window_size
	pattern = pattern.trim_to_idx(start_idx=start_idx, end_idx=end_idx)

	pattern = _expand_seqlets_to_fill_pattern(pattern, 
		track_set=track_set, left_flank_to_add=flank,
		right_flank_to_add=flank)

	return pattern

def _expand_seqlets_to_fill_pattern(pattern, track_set, left_flank_to_add,
	right_flank_to_add):

	new_seqlets = []
	for seqlet in pattern.seqlets:
		left_expansion = left_flank_to_add 
		right_expansion = len(pattern) - len(seqlet) + right_flank_to_add

		if seqlet.is_revcomp == False:
			start = seqlet.start - left_expansion
			end = seqlet.end + right_expansion
		else:
			start = seqlet.start - right_expansion
			end = seqlet.end + left_expansion
		
		if start >= 0 and end <= track_set.length:
			seqlet = track_set.create_seqlets(
				seqlets=[core.Seqlet(example_idx=seqlet.example_idx,
					start=start, end=end, is_revcomp=seqlet.is_revcomp)])[0]

			new_seqlets.append(seqlet)

	if len(new_seqlets) > 0:
		return core.SeqletSet(seqlets=new_seqlets)
	else:
		return None


def _align_patterns(parent_pattern, child_pattern, metric, min_overlap, 
	transformer, include_hypothetical):

	fwd_data_parent, rev_data_parent = util.get_2d_data_from_patterns(
		[parent_pattern], transformer=transformer,
		include_hypothetical=include_hypothetical)

	fwd_data_child, rev_data_child = util.get_2d_data_from_patterns(
		[child_pattern], transformer=transformer,
		include_hypothetical=include_hypothetical)

	best_crossmetric, best_crossmetric_argmax = metric(fwd_data_child, 
		fwd_data_parent, min_overlap).squeeze()

	best_crossmetric_rev, best_crossmetric_argmax_rev = metric(rev_data_child, 
		fwd_data_parent, min_overlap).squeeze()

	if best_crossmetric_rev > best_crossmetric:
		return int(best_crossmetric_argmax_rev), True, best_crossmetric_rev
	else:
		return int(best_crossmetric_argmax), False, best_crossmetric


def merge_in_seqlets_filledges(parent_pattern, seqlets_to_merge,
	track_set, metric, min_overlap, transformer='l1', 
	include_hypothetical=True):

	parent_pattern = parent_pattern.copy()

	for seqlet in seqlets_to_merge:
		alnmt, revcomp_match, alnmt_score = _align_patterns(parent_pattern, 
			seqlet, metric, min_overlap, transformer, include_hypothetical)
		
		if revcomp_match:
			seqlet = seqlet.revcomp()

		preexpansion_seqletlen = len(seqlet)

		left_expansion = max(alnmt,0)
		right_expansion = max((len(parent_pattern) - (alnmt+len(seqlet))), 0)

		if seqlet.is_revcomp == False:
			start = seqlet.start - left_expansion
			end = seqlet.end + right_expansion
		else:
			start = seqlet.start - right_expansion
			end = seqlet.end + left_expansion

		if start >= 0 and end <= track_set.length:
			seqlet = track_set.create_seqlets(
				seqlets=[core.Seqlet(example_idx=seqlet.example_idx,
					start=start, end=end, is_revcomp=seqlet.is_revcomp)])[0] 
		else:
			continue #don't try adding this seqlet

		#also expand the pattern (if needed) so that the seqlet
		# doesn't go over the edge
		parent_left_expansion = max(0, -alnmt)
		parent_right_expansion = max(0, (alnmt+preexpansion_seqletlen)
										 - len(parent_pattern))

		if (parent_left_expansion > 0) or (parent_right_expansion > 0):
			candidate_parent_pattern = _expand_seqlets_to_fill_pattern(
				parent_pattern, track_set=track_set, 
				left_flank_to_add=parent_left_expansion,
				right_flank_to_add=parent_right_expansion)

			if candidate_parent_pattern is not None:
				parent_pattern = candidate_parent_pattern
			else:
				continue

		#add the seqlet in at alignment 0, assuming it's not already
		# part of the pattern
		if seqlet.string not in parent_pattern.unique_seqlets:
			parent_pattern._add_seqlet(seqlet=seqlet)

	return parent_pattern


class PatternMergeHierarchy(object):
	def __init__(self, root_nodes):
		self.root_nodes = root_nodes

	def add_level(self, level_arr):
		self.levels.append(level_arr)

class PatternMergeHierarchyNode(object):

	def __init__(self, pattern, child_nodes=None, parent_node=None,
					   indices_merged=None, submat_crosscontam=None,
					   submat_alignersim=None): 
		self.pattern = pattern 
		if (child_nodes is None):
			child_nodes = []
		self.child_nodes = child_nodes
		self.parent_node = parent_node
		self.indices_merged = indices_merged
		self.submat_crosscontam = submat_crosscontam
		self.submat_alignersim = submat_alignersim


def _detect_spurious_merging(patterns, track_set, perplexity,
	min_in_subcluster, min_overlap, prob_and_pertrack_sim_merge_thresholds,
	prob_and_pertrack_sim_dealbreaker_thresholds,
	min_frac, min_num, flank_to_add, window_size, bg_freq,
	n_seeds, max_seqlets_subsample=1000):

	to_return = []
	for i, pattern in enumerate(patterns):
		if len(pattern.seqlets) > min_in_subcluster:
			pattern.compute_subpatterns(perplexity=perplexity, n_seeds=n_seeds)

			subpatterns = pattern.subcluster_to_subpattern.values()
			refined_subpatterns = SimilarPatternsCollapser(patterns=subpatterns, 
				track_set=track_set, min_overlap=min_overlap, 
				prob_and_pertrack_sim_merge_thresholds=prob_and_pertrack_sim_merge_thresholds,
				prob_and_pertrack_sim_dealbreaker_thresholds=prob_and_pertrack_sim_dealbreaker_thresholds,
				min_frac=min_frac, min_num=min_num, flank_to_add=flank_to_add, window_size=window_size, 
				bg_freq=bg_freq, max_seqlets_subsample=1000)

			to_return.extend(refined_subpatterns[0]) 
		else:
			to_return.append(pattern)
	
	return SimilarPatternsCollapser(patterns=to_return, 
				track_set=track_set, min_overlap=min_overlap, 
				prob_and_pertrack_sim_merge_thresholds=prob_and_pertrack_sim_merge_thresholds,
				prob_and_pertrack_sim_dealbreaker_thresholds=prob_and_pertrack_sim_dealbreaker_thresholds,
				min_frac=min_frac, min_num=min_num, flank_to_add=flank_to_add, window_size=window_size, 
				bg_freq=bg_freq, max_seqlets_subsample=1000)

def SimilarPatternsCollapser(patterns, track_set,
	min_overlap, prob_and_pertrack_sim_merge_thresholds,
	prob_and_pertrack_sim_dealbreaker_thresholds,
	min_frac, min_num, flank_to_add, window_size, bg_freq,
	max_seqlets_subsample=1000):
	patterns = [x.copy() for x in patterns]

	merge_hierarchy_levels = []        
	current_level_nodes = [
		PatternMergeHierarchyNode(pattern=x) for x in patterns]
	merge_hierarchy_levels.append(current_level_nodes)

	merge_occurred_last_iteration = True

	#negative numbers to indicate which
	# entries need to be filled (versus entries we can infer
	# from the previous iteration of the while loop)
	pairwise_aurocs = -np.ones((len(patterns), len(patterns)))
	pairwise_sims = np.zeros((len(patterns), len(patterns)))

	#loop until no more patterns get merged
	while merge_occurred_last_iteration:
		merge_occurred_last_iteration = False

		#Let's subsample 'patterns' to prevent runtime from being too
		# large in calculating pairwise sims. 
		subsample_patterns = []
		for pattern in patterns:
			if len(pattern.seqlets) > max_seqlets_subsample:
				subsample = np.random.RandomState(1234).choice(
					a=pattern.seqlets, replace=False, size=max_seqlets_subsample)
				pattern = core.SeqletSet(seqlets=subsample)

			subsample_patterns.append(pattern)

		n = len(patterns)
		for i in range(n):
			for j in range(n):
				#Note: I compute both i,j AND j,i because although
				# the result is the same for the sim, it can be different
				# for the auroc because a different motif is getting
				# shifted over.
				if j == i:
					pairwise_aurocs[i, j] = 0.5
					pairwise_sims[i, j] = 1.0
					continue

				if pairwise_aurocs[i, j] >= 0: #filled in from previous iter
					continue 

				#Compute best alignment between pattern pair
				alnmt, rc, aligner_sim =\
					_align_patterns(parent_pattern=patterns[i], 
						child_pattern=patterns[j], 
						metric=affinitymat.pearson_correlation, 
						min_overlap=min_overlap, 
						include_hypothetical=False,
						transformer='magnitude') 

				pairwise_sims[i, j] = aligner_sim

				#get realigned pattern2
				pattern2_coords = subsample_patterns[j].seqlets
				if rc: #flip strand if needed to align
					pattern2_coords  = [x.revcomp() for x in pattern2_coords]

				#now apply the alignment
				pattern2_coords = [x.shift((1 if x.is_revcomp else -1)*alnmt)
					for x in pattern2_coords] 

				# Filter out bad seqlets
				pattern2_coords = [seqlet for seqlet in pattern2_coords 
					if seqlet.start >= 0 and seqlet.end < track_set.length]

				if len(pattern2_coords) == 0:
					pairwise_sims[i, j] = 0.0
					pairwise_aurocs[i, j] = 0.5
					continue

				pattern2_shifted_seqlets = track_set.create_seqlets(
					seqlets=pattern2_coords)

				pattern1_fwdseqdata, _ =\
				  util.get_2d_data_from_patterns(subsample_patterns[i].seqlets)

				pattern2_fwdseqdata, _ =\
				  util.get_2d_data_from_patterns(pattern2_shifted_seqlets)

				#Flatten, compute continjacc sim at this alignment
				flat_pattern1_fwdseqdata = pattern1_fwdseqdata.reshape(
					(len(pattern1_fwdseqdata), -1))
				flat_pattern2_fwdseqdata = pattern2_fwdseqdata.reshape(
					(len(pattern2_fwdseqdata), -1))

				between_pattern_sims = affinitymat.jaccard(
					flat_pattern1_fwdseqdata[:, :, None], 
					flat_pattern2_fwdseqdata[:, :, None])[:, :, 0].flatten()

				within_pattern1_sims = affinitymat.jaccard(
					flat_pattern1_fwdseqdata[:, :, None], 
					flat_pattern1_fwdseqdata[:, :, None])[:, :, 0].flatten()

				auroc = roc_auc_score(
					y_true=[0 for x in between_pattern_sims]
						   +[1 for x in within_pattern1_sims],
					y_score=list(between_pattern_sims)
							+list(within_pattern1_sims))

				#The symmetrization over i,j and j,i is done later
				pairwise_aurocs[i,j] = auroc


		#pairwise_sims is not symmetric; differ based on which pattern is
		# padded with zeros.
		patterns_to_patterns_aligner_sim = 0.5*(pairwise_sims+pairwise_sims.T)
		cross_contamination = 2*(1-np.maximum(pairwise_aurocs,0.5))

		indices_to_merge = []
		merge_partners_so_far = dict([(i, set([i])) for i in
									  range(len(patterns))])

		#merge patterns with highest similarity first
		sorted_pairs = sorted([(i,j,patterns_to_patterns_aligner_sim[i,j])
						for i in range(len(patterns))
						for j in range(len(patterns)) if (i < j)],
						key=lambda x: -x[2])

		#iterate over pairs
		for i, j, aligner_sim in sorted_pairs:
			#symmetrize asymmetric crosscontam
			# take min rather than avg to avoid aggressive merging
			cross_contam = min(cross_contamination[i,j],
								cross_contamination[j,i])

			collapse = any([(cross_contam >= x[0] and aligner_sim >= x[1])
				for x in prob_and_pertrack_sim_merge_thresholds])

			if collapse:
				collapse_passed = True
				#check compatibility for all indices that are
				#about to be merged
				merge_under_consideration = set(
					list(merge_partners_so_far[i])
					+list(merge_partners_so_far[j]))

				for m1 in merge_under_consideration:
					for m2 in merge_under_consideration:
						if (m1 < m2):
							cross_contam_here =\
								0.5*(cross_contamination[m1, m2]+
									 cross_contamination[m2, m1])
							aligner_sim_here =\
								patterns_to_patterns_aligner_sim[
									m1, m2]

							dealbreaker = any([(cross_contam_here <= x[0] and aligner_sim_here <= x[1])              
								for x in prob_and_pertrack_sim_dealbreaker_thresholds])

							if dealbreaker:
								collapse_passed = False
								break

				if collapse_passed:
					indices_to_merge.append((i,j))
					for an_idx in merge_under_consideration:
						merge_partners_so_far[an_idx]=\
							merge_under_consideration 

		for i, j in indices_to_merge:
			pattern1 = patterns[i]
			pattern2 = patterns[j]

			if pattern1 != pattern2: #if not the same object
				if len(pattern1.seqlets) < len(pattern2.seqlets):
					parent_pattern, child_pattern = pattern2, pattern1
				else:
					parent_pattern, child_pattern = pattern1, pattern2

				new_pattern = merge_in_seqlets_filledges(
					parent_pattern=parent_pattern,
					seqlets_to_merge=child_pattern.seqlets,
					include_hypothetical=False,
					metric=affinitymat.pearson_correlation,
					min_overlap=min_overlap,
					transformer='magnitude',
					track_set=track_set)

				new_pattern = polish_pattern(new_pattern, min_frac=min_frac, 
					min_num=min_num, track_set=track_set, flank=flank_to_add, 
					window_size=window_size, bg_freq=bg_freq)

				if new_pattern is not None:
					for k in range(len(patterns)):
						#Replace EVERY case where the parent or child
						# pattern is present with the new pattern. This
						# effectively does single-linkage.
						if (patterns[k]==parent_pattern or
							patterns[k]==child_pattern):
							patterns[k]=new_pattern

		merge_occurred_last_iteration = (len(indices_to_merge) > 0)

		if merge_occurred_last_iteration:
			#Once we are here, each element of 'patterns'
			#will have the new parent of the corresponding element
			#of 'old_patterns'
			old_to_new_pattern_mapping = patterns

			#sort by size and remove redundant patterns
			patterns = sorted(patterns, key=lambda x: -len(x.seqlets))
			patterns = list(OrderedDict([(x,1) for x in patterns]).keys())

			#let's figure out which indices don't require recomputation
			# and use it to repopulate pairwise_sims and pairwise_aurocs
			old_to_new_index_mappings = OrderedDict()
			for old_pattern_idx,(old_pattern_node, corresp_new_pattern)\
				in enumerate(zip(current_level_nodes,
								 old_to_new_pattern_mapping)):
				#if the old pattern was NOT changed in this iteration
				if old_pattern_node.pattern == corresp_new_pattern:
					new_idx = patterns.index(corresp_new_pattern) 
					old_to_new_index_mappings[old_pattern_idx] = new_idx

			new_pairwise_aurocs = -np.ones((len(patterns), len(patterns)))
			new_pairwise_sims = np.zeros((len(patterns), len(patterns)))
			for old_idx_i, new_idx_i in\
				old_to_new_index_mappings.items():
				for old_idx_j, new_idx_j in\
					old_to_new_index_mappings.items():
					new_pairwise_aurocs[new_idx_i, new_idx_j] =\
						pairwise_aurocs[old_idx_i, old_idx_j]
					new_pairwise_sims[new_idx_i, new_idx_j] =\
						pairwise_sims[old_idx_i, old_idx_j]
			pairwise_aurocs = new_pairwise_aurocs 
			pairwise_sims = new_pairwise_sims
				 

			#update the hierarchy
			#the current 'top level' will consist of all the current
			# nodes that didn't get a new parent, plus any new parents
			# created                
			next_level_nodes = []
			for frontier_pattern in patterns:
				#either this pattern is in old_pattern_nodes, in which
				# case take the old_pattern_node entry, or it's a completely
				# new pattern in which case make a node for it
				old_pattern_node_found = False
				for old_pattern_node in current_level_nodes:
					if (old_pattern_node.pattern==frontier_pattern):
						#sanity check..there should be only one node
						# per pattern
						next_level_nodes.append(old_pattern_node)
						old_pattern_node_found = True 
				if (old_pattern_node_found==False):
				   next_level_nodes.append(
					PatternMergeHierarchyNode(frontier_pattern)) 

			for next_level_node in next_level_nodes:
				#iterate over all the old patterns and their new parent
				# in order to set up the child nodes correctly
				for old_pattern_idx,(old_pattern_node, corresp_new_pattern)\
					in enumerate(zip(current_level_nodes,
									 old_to_new_pattern_mapping)):

					#if the node has a new parent
					if (old_pattern_node.pattern != corresp_new_pattern):
						if (next_level_node.pattern==corresp_new_pattern):

							
							#corresp_new_pattern should be comprised of a 
							# merging of all the old patterns at
							# indices_merged_with
							indices_merged = tuple(sorted(
								merge_partners_so_far[old_pattern_idx])) 
							#get the relevant slice         
							submat_crosscontam =\
							 cross_contamination[indices_merged,:][:,
												 indices_merged]
							submat_alignersim =\
							 patterns_to_patterns_aligner_sim[
								indices_merged, :][:,indices_merged]

							if (next_level_node.indices_merged is not None):
								assert (next_level_node.indices_merged
										==indices_merged),\
								 (next_level_node.indices_merged,
								  indices_merged)
							else:
								next_level_node.indices_merged =\
									indices_merged
								next_level_node.submat_crosscontam =\
									submat_crosscontam
								next_level_node.submat_alignersim =\
									submat_alignersim

							next_level_node.child_nodes.append(
											old_pattern_node) 

							old_pattern_node.parent_node = next_level_node

			current_level_nodes = next_level_nodes

	return patterns, PatternMergeHierarchy(root_nodes=current_level_nodes)
	


# gapped_kmer.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import scipy
import numpy as np

from numba import njit, prange
import numba

key_type = numba.types.int64
value_type = numba.types.float64

@njit(parallel=True)
def _extract_gkmers(X, min_k, max_k, max_gap, max_len, max_entries):
	nx = X.shape[0]
	keys = np.zeros((nx, max_entries), dtype='int64')
	scores = np.zeros((nx, max_entries), dtype='float64')

	for xi in prange(nx):
		n = X.shape[1]
		gkmer_attrs = numba.typed.Dict.empty(key_type=key_type, value_type=value_type)

		last_k_gkmers = []
		last_k_gkmers_attrs = []
		last_k_gkmers_hashes = []

		for i in range(n):
			base = int(X[xi, i, 1])
			attr = X[xi, i, 2]

			last_k_gkmers.append(np.array([i], dtype='int32'))
			last_k_gkmers_attrs.append(np.array([attr], dtype='float64'))
			last_k_gkmers_hashes.append(np.array([base+1], dtype='int64'))

		for k in range(2, max_k+1):
			for j in range(n):
				start_position = X[xi, j, 0]

				gkmers_ = []
				gkmer_attrs_ = []
				gkmer_hashes_ = []

				for i in range(j+1, n):
					position = X[xi, i, 0]
					base = int(X[xi, i, 1])
					attr = X[xi, i, 2]

					if (position - start_position) >= max_len:
						break

					for g in range(len(last_k_gkmers[j])):
						gkmer = last_k_gkmers[j][g]
						gkmer_attr = last_k_gkmers_attrs[j][g]
						gkmer_hash = last_k_gkmers_hashes[j][g]

						last_position = X[xi, gkmer, 0]
						if last_position >= position:
							break

						if (position - last_position) > max_gap:
							continue

						diff = int(position - last_position - 1)
						length = int(position - start_position)

						new_gkmer_hash = gkmer_hash + (base+1) * (5 ** length)
						new_gkmer_attr = gkmer_attr + attr
						
						gkmers_.append(i)
						gkmer_attrs_.append(new_gkmer_attr)
						gkmer_hashes_.append(new_gkmer_hash)

						if k >= min_k:
							gkmer_attrs[new_gkmer_hash] = gkmer_attrs.get(new_gkmer_hash, 0) + new_gkmer_attr / k

				if len(gkmers_) == 0:
					last_k_gkmers[j] = np.zeros(0, dtype='int32')
					last_k_gkmers_attrs[j] = np.zeros(0, dtype='float64')
					last_k_gkmers_hashes[j] = np.zeros(0, dtype='int64')
				else:
					last_k_gkmers[j] = np.array(gkmers_, dtype='int32')
					last_k_gkmers_attrs[j] = np.array(gkmer_attrs_, dtype='float64')
					last_k_gkmers_hashes[j] = np.array(gkmer_hashes_, dtype='int64')
		
		ny = len(gkmer_attrs)
		keys_ = np.empty(ny, dtype='int64')
		scores_ = np.empty(ny, dtype='float64')

		for i, key in enumerate(gkmer_attrs.keys()):
			keys_[i] = key
			scores_[i] = gkmer_attrs[key]

		idxs = np.argsort(-np.abs(scores_), kind='mergesort')[:max_entries]

		keys[xi] = keys_[idxs]
		scores[xi] = scores_[idxs]

	return keys, scores

def _seqlet_to_gkmers(seqlets, topn, min_k, max_k, max_gap, max_len, 
	max_entries, take_fwd, sign):

	Xs = []
	for seqlet in seqlets:
		onehot = seqlet.sequence
		contrib_scores = seqlet.hypothetical_contribs*onehot*sign

		if not take_fwd:
			onehot = onehot[::-1, ::-1]
			contrib_scores = contrib_scores[::-1, ::-1]

		#get the top n positiosn
		per_pos_imp = np.sum(contrib_scores, axis=-1)
		per_pos_bases = np.argmax(onehot, axis=-1)

		#get the top n positions
		topn_pos = np.argsort(-per_pos_imp)[:topn]

		X_ = sorted([(pos, per_pos_bases[pos], per_pos_imp[pos]) for pos in topn_pos], key=lambda x:x[0])
		X_ = np.array(X_)
		Xs.append(X_)

	X = np.array(Xs)
	keys, scores = _extract_gkmers(X, min_k=min_k, max_k=max_k, 
		max_gap=max_gap, max_len=max_len, max_entries=max_entries)
	
	row_idxs = np.repeat(range(keys.shape[0]), keys.shape[1])
	csr_mat = scipy.sparse.csr_matrix((scores.flatten(), 
		(row_idxs, keys.flatten())), shape=(len(keys), 5**max_len))

	return csr_mat



from setuptools import setup

setup(
	name='modisco-lite',
	version='2.0.7',
	author='Jacob Schreiber',
	author_email='jmschreiber91@gmail.com',
	packages=['modiscolite'],
	scripts=['modisco'],
	url='https://github.com/jmschrei/tfmodisco-lite',
	license='LICENSE.txt',
	description='Transcription Factor MOtif Discovery from Importance SCOres - lite',
	install_requires=[
		'numpy >= 1.21.5', 
		'scipy >= 1.6.2',
		'numba >= 0.53.1',
		'scikit-learn >= 1.0.2',
		'leidenalg == 0.8.10',
		'igraph == 0.9.11',
		'tqdm >= 4.38.0',
		'pandas >= 1.4.3',
		'logomaker >= 0.8',
		'h5py >= 3.7.0',
		'hdf5plugin'
	]
)



# cluster.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import leidenalg
import numpy as np
import igraph as ig

def LeidenCluster(affinity_mat, n_seeds=2, n_leiden_iterations=-1):
    n_vertices = affinity_mat.shape[0]
    n_cols = affinity_mat.indptr
    sources = np.concatenate([np.ones(n_cols[i+1] - n_cols[i], dtype='int32') * i for i in range(n_vertices)])

    g = ig.Graph(directed=None) 
    g.add_vertices(n_vertices)
    g.add_edges(zip(sources, affinity_mat.indices))

    best_clustering = None
    best_quality = None

    for seed in range(1, n_seeds+1):
        partition = leidenalg.find_partition(
            graph=g,
            partition_type=leidenalg.ModularityVertexPartition,
            weights=affinity_mat.data,
            n_iterations=n_leiden_iterations,
            initial_membership=None,
            seed=seed*100) 

        quality = np.array(partition.quality())
        membership = np.array(partition.membership)
        
        if best_quality is None or quality > best_quality:
            best_quality = quality
            best_clustering = membership

    return best_clustering



# tfmodisco.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import numpy as np

import scipy
import scipy.sparse

from collections import OrderedDict
from collections import defaultdict

from . import affinitymat
from . import aggregator
from . import extract_seqlets
from . import core
from . import util
from . import cluster

def _density_adaptation(affmat_nn, seqlet_neighbors, tsne_perplexity):
	eps = 0.0000001

	rows, cols, data = [], [], []
	for row in range(len(affmat_nn)):
		for col, datum in zip(seqlet_neighbors[row], affmat_nn[row]):
			rows.append(row)
			cols.append(col)
			data.append(datum)

	affmat_nn = scipy.sparse.csr_matrix((data, (rows, cols)), 
		shape=(len(affmat_nn), len(affmat_nn)), dtype='float64')
	
	affmat_nn.data = np.maximum(np.log((1.0/(0.5*np.maximum(affmat_nn.data, eps)))-1), 0)
	affmat_nn.eliminate_zeros()

	counts_nn = scipy.sparse.csr_matrix((np.ones_like(affmat_nn.data), 
		affmat_nn.indices, affmat_nn.indptr), shape=affmat_nn.shape, dtype='float64')

	affmat_nn += affmat_nn.T
	counts_nn += counts_nn.T
	affmat_nn.data /= counts_nn.data
	del counts_nn

	betas = [util.binary_search_perplexity(tsne_perplexity, affmat_nn[i].data) for i in range(affmat_nn.shape[0])]
	normfactors = np.array([np.exp(-np.array(affmat_nn[i].data)/beta).sum()+1 for i, beta in enumerate(betas)])

	for i in range(affmat_nn.shape[0]):
		for j_idx in range(affmat_nn.indptr[i], affmat_nn.indptr[i+1]):
			j = affmat_nn.indices[j_idx]
			distance = affmat_nn.data[j_idx]

			rbf_i = np.exp(-distance / betas[i]) / normfactors[i]
			rbf_j = np.exp(-distance / betas[j]) / normfactors[j]

			affmat_nn.data[j_idx] = np.sqrt(rbf_i * rbf_j)

	affmat_diags = scipy.sparse.diags(1.0 / normfactors)
	affmat_nn += affmat_diags
	return affmat_nn


def _filter_patterns(patterns, min_seqlet_support, window_size, 
	min_ic_in_window, background, ppm_pseudocount):
	passing_patterns = []
	for pattern in patterns:
		if len(pattern.seqlets) < min_seqlet_support:
			continue

		ppm = pattern.sequence
		per_position_ic = util.compute_per_position_ic(ppm=ppm, 
			background=background, pseudocount=ppm_pseudocount)

		if len(per_position_ic) < window_size:       
			if np.sum(per_position_ic) < min_ic_in_window:
				continue
		else:
			#do the sliding window sum rearrangement
			windowed_ic = np.sum(util.rolling_window(
				a=per_position_ic, window=window_size), axis=-1)

			if np.max(windowed_ic) < min_ic_in_window:
				continue

		passing_patterns.append(pattern)

	return passing_patterns


def _patterns_from_clusters(seqlets, track_set, min_overlap,
	min_frac, min_num, flank_to_add, window_size, bg_freq, cluster_indices, 
	track_sign):

	seqlet_sort_metric = lambda x: -np.sum(np.abs(x.contrib_scores))
	num_clusters = max(cluster_indices+1)
	cluster_to_seqlets = defaultdict(list) 

	for seqlet, idx in zip(seqlets, cluster_indices):
		cluster_to_seqlets[idx].append(seqlet)

	patterns = []
	for i in range(num_clusters):
		sorted_seqlets = sorted(cluster_to_seqlets[i], key=seqlet_sort_metric) 
		pattern = core.SeqletSet([sorted_seqlets[0]])

		if len(sorted_seqlets) > 1:
			pattern = aggregator.merge_in_seqlets_filledges(
				parent_pattern=pattern, seqlets_to_merge=sorted_seqlets[1:],
				track_set=track_set, metric=affinitymat.jaccard,
				min_overlap=min_overlap)

		pattern = aggregator.polish_pattern(pattern, min_frac=min_frac, 
			min_num=min_num, track_set=track_set, flank=flank_to_add, 
			window_size=window_size, bg_freq=bg_freq)

		if pattern is not None:
			if np.sign(np.sum(pattern.contrib_scores)) == track_sign:
				patterns.append(pattern)

	return patterns


def _filter_by_correlation(seqlets, seqlet_neighbors, coarse_affmat_nn, 
	fine_affmat_nn, correlation_threshold):

	correlations = []
	for fine_affmat_row, coarse_affmat_row in zip(fine_affmat_nn, coarse_affmat_nn):
		to_compare_mask = np.abs(fine_affmat_row) > 0
		corr = scipy.stats.spearmanr(fine_affmat_row[to_compare_mask],
			coarse_affmat_row[to_compare_mask])
		correlations.append(corr.correlation)

	correlations = np.array(correlations)
	filtered_rows_mask = np.array(correlations) > correlation_threshold

	filtered_seqlets = [seqlet for seqlet, mask in zip(seqlets, 
		filtered_rows_mask) if mask == True]

	#figure out a mapping from pre-filtering to the
	# post-filtering indices
	new_idx_mapping = np.cumsum(filtered_rows_mask) - 1
	retained_indices = set(np.where(filtered_rows_mask == True)[0])

	filtered_neighbors = []
	filtered_affmat_nn = []
	for old_row_idx, (old_neighbors, affmat_row) in enumerate(zip(seqlet_neighbors, fine_affmat_nn)):
		if old_row_idx in retained_indices:
			filtered_old_neighbors = [neighbor for neighbor in old_neighbors if neighbor in retained_indices]
			filtered_affmat_row = [affmatval for affmatval, neighbor in zip(affmat_row,old_neighbors) if neighbor in retained_indices]
			filtered_neighbors_row = [new_idx_mapping[neighbor] for neighbor in filtered_old_neighbors]
			filtered_neighbors.append(filtered_neighbors_row)
			filtered_affmat_nn.append(filtered_affmat_row)

	return filtered_seqlets, filtered_neighbors, filtered_affmat_nn


def seqlets_to_patterns(seqlets, track_set, track_signs=None, 
	min_overlap_while_sliding=0.7, nearest_neighbors_to_compute=500, 
	affmat_correlation_threshold=0.15, tsne_perplexity=10.0, 
	n_leiden_iterations=-1, n_leiden_runs=50, frac_support_to_trim_to=0.2,
	min_num_to_trim_to=30, trim_to_window_size=20, initial_flank_to_add=5,
	prob_and_pertrack_sim_merge_thresholds=[(0.8,0.8), (0.5, 0.85), (0.2, 0.9)],
	prob_and_pertrack_sim_dealbreaker_thresholds=[(0.4, 0.75), (0.2,0.8), (0.1, 0.85), (0.0,0.9)],
	subcluster_perplexity=50, merging_max_seqlets_subsample=300,
	final_min_cluster_size=20,min_ic_in_window=0.6, min_ic_windowsize=6,
	ppm_pseudocount=0.001):

	bg_freq = np.mean([seqlet.sequence for seqlet in seqlets], axis=(0, 1)) 

	seqlets_sorter = (lambda arr: sorted(arr, key=lambda x:
		-np.sum(np.abs(x.contrib_scores))))

	seqlets = seqlets_sorter(seqlets)

	for round_idx in range(2):
		if len(seqlets) == 0:
			return None

		# Step 1: Generate coarse resolution
		coarse_affmat_nn, seqlet_neighbors = affinitymat.cosine_similarity_from_seqlets(
			seqlets=seqlets, n_neighbors=nearest_neighbors_to_compute, sign=track_signs)

		# Step 2: Generate fine representation
		fine_affmat_nn = affinitymat.jaccard_from_seqlets(
			seqlets=seqlets, seqlet_neighbors=seqlet_neighbors,
			min_overlap=min_overlap_while_sliding)

		if round_idx == 0:
			filtered_seqlets, seqlet_neighbors, filtered_affmat_nn = (
				_filter_by_correlation(seqlets, seqlet_neighbors, 
					coarse_affmat_nn, fine_affmat_nn, 
					affmat_correlation_threshold))
		else:
			filtered_seqlets = seqlets
			filtered_affmat_nn = fine_affmat_nn

		del coarse_affmat_nn
		del fine_affmat_nn
		del seqlets

		# Step 4: Density adaptation
		csr_density_adapted_affmat = _density_adaptation(
			filtered_affmat_nn, seqlet_neighbors, tsne_perplexity)

		del filtered_affmat_nn
		del seqlet_neighbors

		# Step 5: Clustering
		cluster_indices = cluster.LeidenCluster(
			csr_density_adapted_affmat,
			n_seeds=n_leiden_runs,
			n_leiden_iterations=n_leiden_iterations)

		del csr_density_adapted_affmat

		patterns = _patterns_from_clusters(filtered_seqlets, 
			track_set=track_set, 
			min_overlap=min_overlap_while_sliding, 
			min_frac=frac_support_to_trim_to, 
			min_num=min_num_to_trim_to, 
			flank_to_add=initial_flank_to_add, 
			window_size=trim_to_window_size, 
			bg_freq=bg_freq, 
			cluster_indices=cluster_indices, 
			track_sign=track_signs)

		#obtain unique seqlets from adjusted motifs
		seqlets = list(dict([(y.string, y)
						 for x in patterns for y in x.seqlets]).values())

	del seqlets

	merged_patterns, pattern_merge_hierarchy = aggregator._detect_spurious_merging(
		patterns=patterns, track_set=track_set, perplexity=subcluster_perplexity, 
		min_in_subcluster=max(final_min_cluster_size, subcluster_perplexity), 
		min_overlap=min_overlap_while_sliding,
		prob_and_pertrack_sim_merge_thresholds=prob_and_pertrack_sim_merge_thresholds,
		prob_and_pertrack_sim_dealbreaker_thresholds=prob_and_pertrack_sim_dealbreaker_thresholds,
		min_frac=frac_support_to_trim_to, min_num=min_num_to_trim_to,
		flank_to_add=initial_flank_to_add,
		window_size=trim_to_window_size, bg_freq=bg_freq,
		max_seqlets_subsample=merging_max_seqlets_subsample,
		n_seeds=n_leiden_runs)

	#Now start merging patterns 
	merged_patterns = sorted(merged_patterns, key=lambda x: -len(x.seqlets))

	patterns = _filter_patterns(merged_patterns, 
		min_seqlet_support=final_min_cluster_size, 
		window_size=min_ic_windowsize, min_ic_in_window=min_ic_in_window, 
		background=bg_freq, ppm_pseudocount=ppm_pseudocount)

	#apply subclustering procedure on the final patterns
	for patternidx, pattern in enumerate(patterns):
		pattern.compute_subpatterns(subcluster_perplexity, 
			n_seeds=n_leiden_runs, n_iterations=n_leiden_iterations)

	return patterns


def TFMoDISco(one_hot, hypothetical_contribs, sliding_window_size=21, 
	flank_size=10, min_metacluster_size=100,
	weak_threshold_for_counting_sign=0.8, max_seqlets_per_metacluster=20000,
	target_seqlet_fdr=0.2, min_passing_windows_frac=0.03,
	max_passing_windows_frac=0.2, n_leiden_runs=50, n_leiden_iterations=-1, 
	min_overlap_while_sliding=0.7, nearest_neighbors_to_compute=500, 
	affmat_correlation_threshold=0.15, tsne_perplexity=10.0, 
	frac_support_to_trim_to=0.2, min_num_to_trim_to=30, trim_to_window_size=20, 
	initial_flank_to_add=5,
	prob_and_pertrack_sim_merge_thresholds=[(0.8,0.8), (0.5, 0.85), (0.2, 0.9)],
	prob_and_pertrack_sim_dealbreaker_thresholds=[(0.4, 0.75), (0.2,0.8), (0.1, 0.85), (0.0,0.9)],
	subcluster_perplexity=50, merging_max_seqlets_subsample=300,
	final_min_cluster_size=20, min_ic_in_window=0.6, min_ic_windowsize=6,
	ppm_pseudocount=0.001, verbose=False):

	contrib_scores = np.multiply(one_hot, hypothetical_contribs)

	track_set = core.TrackSet(one_hot=one_hot, 
		contrib_scores=contrib_scores,
		hypothetical_contribs=hypothetical_contribs)

	seqlet_coords, threshold = extract_seqlets.extract_seqlets(
		attribution_scores=contrib_scores.sum(axis=2),
		window_size=sliding_window_size,
		flank=flank_size,
		suppress=(int(0.5*sliding_window_size) + flank_size),
		target_fdr=target_seqlet_fdr,
		min_passing_windows_frac=min_passing_windows_frac,
		max_passing_windows_frac=max_passing_windows_frac,
		weak_threshold_for_counting_sign=weak_threshold_for_counting_sign) 

	seqlets = track_set.create_seqlets(seqlet_coords) 

	pos_seqlets, neg_seqlets = [], []
	for seqlet in seqlets:
		flank = int(0.5*(len(seqlet)-sliding_window_size))
		attr = np.sum(seqlet.contrib_scores[flank:-flank])

		if attr > threshold:
			pos_seqlets.append(seqlet)
		elif attr < -threshold:
			neg_seqlets.append(seqlet)

	del seqlets

	if len(pos_seqlets) > min_metacluster_size:
		pos_seqlets = pos_seqlets[:max_seqlets_per_metacluster]
		if verbose:
			print("Using {} positive seqlets".format(len(pos_seqlets)))

		pos_patterns = seqlets_to_patterns(seqlets=pos_seqlets,
			track_set=track_set, 
			track_signs=1,
			min_overlap_while_sliding=min_overlap_while_sliding,
			nearest_neighbors_to_compute=nearest_neighbors_to_compute,
			affmat_correlation_threshold=affmat_correlation_threshold,
			tsne_perplexity=tsne_perplexity,
			n_leiden_iterations=n_leiden_iterations,
			n_leiden_runs=n_leiden_runs,
			frac_support_to_trim_to=frac_support_to_trim_to,
			min_num_to_trim_to=min_num_to_trim_to,
			trim_to_window_size=trim_to_window_size,
			initial_flank_to_add=initial_flank_to_add,
			prob_and_pertrack_sim_merge_thresholds=prob_and_pertrack_sim_merge_thresholds,
			prob_and_pertrack_sim_dealbreaker_thresholds=prob_and_pertrack_sim_dealbreaker_thresholds,
			subcluster_perplexity=subcluster_perplexity,
			merging_max_seqlets_subsample=merging_max_seqlets_subsample,
			final_min_cluster_size=final_min_cluster_size,
			min_ic_in_window=min_ic_in_window,
			min_ic_windowsize=min_ic_windowsize,
			ppm_pseudocount=ppm_pseudocount)
	else:
		pos_patterns = None

	if len(neg_seqlets) > min_metacluster_size:
		neg_seqlets = neg_seqlets[:max_seqlets_per_metacluster]
		if verbose:
			print("Extracted {} negative seqlets".format(len(neg_seqlets)))

		neg_patterns = seqlets_to_patterns(seqlets=neg_seqlets,
			track_set=track_set, 
			track_signs=-1,
			min_overlap_while_sliding=min_overlap_while_sliding,
			nearest_neighbors_to_compute=nearest_neighbors_to_compute,
			affmat_correlation_threshold=affmat_correlation_threshold,
			tsne_perplexity=tsne_perplexity,
			n_leiden_iterations=n_leiden_iterations,
			n_leiden_runs=n_leiden_runs,
			frac_support_to_trim_to=frac_support_to_trim_to,
			min_num_to_trim_to=min_num_to_trim_to,
			trim_to_window_size=trim_to_window_size,
			initial_flank_to_add=initial_flank_to_add,
			prob_and_pertrack_sim_merge_thresholds=prob_and_pertrack_sim_merge_thresholds,
			prob_and_pertrack_sim_dealbreaker_thresholds=prob_and_pertrack_sim_dealbreaker_thresholds,
			subcluster_perplexity=subcluster_perplexity,
			merging_max_seqlets_subsample=merging_max_seqlets_subsample,
			final_min_cluster_size=final_min_cluster_size,
			min_ic_in_window=min_ic_in_window,
			min_ic_windowsize=min_ic_windowsize,
			ppm_pseudocount=ppm_pseudocount)
	else:
		neg_patterns = None

	return pos_patterns, neg_patterns
